import pytest
from pytest import param

from transport_data.oica import convert, get_cl_geo, update_registry
from transport_data.testing import MARK


@pytest.mark.parametrize(
    "measure, N_obs",
    (
        param("PROD", None, marks=pytest.mark.xfail(raises=NotImplementedError)),
        param("SALES", dict(SALES=1332, SALES_GR=444)),
        param(
            "STOCK", dict(STOCK=462, STOCK_AAGR=230, STOCK_CAP=77), marks=MARK["#52"]
        ),
        param(
            "STOCK_AAGR",
            dict(STOCK=462, STOCK_AAGR=230, STOCK_CAP=77),
            marks=MARK["#52"],
        ),
    ),
)
def test_convert(measure: str, N_obs: dict[str, int]) -> None:
    result = convert(measure)

    # GEO codelist written with some elements
    assert 0 < len(get_cl_geo())

    # Results contain the expected number of data sets
    assert len(N_obs) == len(result)

    # Each data set contains the expected number of observations
    observed = {dfd_id.split("DF_OICA_")[-1]: len(ds) for dfd_id, ds in result.items()}
    assert N_obs == observed

    # DEBUG
    # import sdmx
    #
    # for _, ds in result.items():
    #     print(sdmx.to_pandas(ds).to_string())


def test_update_registry() -> None:
    update_registry()



"""Local data storage.

.. todo:: Add the following to base capabilities of :class:`.UnionStore`::

   - :py:`write(..., annotate=True)`
"""

import logging
from typing import TYPE_CHECKING

import click
import dsss.store
import sdmx.model.v21 as m
import sdmx.urn

from transport_data.util.sdmx import anno_generated

if TYPE_CHECKING:
    import transport_data.config

log = logging.getLogger(__name__)


class UnionStore(dsss.store.UnionStore):
    """A store that maps between a :class:`.LocalStore` and :class:`.Registry`.

    This class extends :class:`dsss.store.UnionStore` with the following behaviour:

    - The sub-stores are initialized from an instance of :class:`.Config`.
    - :meth:`assign_version` and :meth:`get` have extended behaviour. See the
      documentation of each method.
    - Additional methods :meth:`clone` and :meth:`add_to_registry`.
    """

    def __init__(self, config: "transport_data.config.Config") -> None:
        super().__init__(
            hook={"before set": anno_generated},
            store=dict(
                local=dsss.store.FlatFileStore(path=config.data_path.joinpath("local")),
                registry=dsss.store.GitStore(
                    path=config.data_path.joinpath("registry"),
                    remote_url=config.registry_remote_url,
                ),
            ),
        )
        # FIXME Include in the above call, once dsss.store.UnionStore supports
        self.maintainer_store.update(ISO="registry")

    # Overrides of methods of dsss.store.UnionStore or its parents

    def get(self, key: str):
        """Return an object given its `key`.

        Compared to the parent class, this implementation:

        - Expands `key` from a partial URN (e.g. "AgencyScheme=TDCI:TDCI(1.0.0)") to a
          full URN.
        - Handles `key` that is a partial URN without a specific version, e.g.
          "AgencyScheme=TDCI:TDCI". In this case, the greatest version of the referenced
          artefact is returned.
        """
        try:
            full_urn = sdmx.urn.normalize(sdmx.urn.expand(key))
            return super().get(full_urn)
        except KeyError:
            match = sdmx.urn.match(full_urn)
            if match and match["version"] in (None, "None"):
                # No version specified → iterate over all versions
                klass = m.get_class(match["class"])
                assert klass is not None
                if versions := self.list_versions(klass, match["agency"], match["id"]):
                    return super().get(full_urn.replace("(None)", f"({versions[-1]})"))
            raise

    # Additional methods for this class

    def add_to_registry(self, urn: str):
        """Copy data for `urn` from :class:`.LocalStore` to the :class:`.Registry`."""
        full_urn = sdmx.urn.expand(urn)
        obj = self.store["local"].get(full_urn)
        self.store["registry"].set(obj)

    # FIXME Restore this original once the upstream issue is resolved
    # def clone(self):
    #     """Clone the underlying :class:`.Registry`."""
    #     self.store["registry"].clone()

    def clone(self):
        """Clone the underlying :class:`.Registry`.

        Temporary version to work around `khaeru/dsss#19
        <https://github.com/khaeru/dsss/issues/19>`_. This is identical to
        :meth:`dsss.store.GitStore.clone`, except only certain paths are checked out.
        """
        import git

        s = self.store["registry"]

        # Ensure there is a remote for the origin
        try:
            s.repo.delete_remote("origin")
        except git.exc.GitCommandError:
            pass

        origin = s.repo.create_remote("origin", s.remote_url)

        # Fetch the remote
        branch_name = "main"
        origin.fetch(f"refs/heads/{branch_name}")
        b = origin.refs[branch_name]

        # Check out the branch
        try:
            head = s.repo.heads[branch_name]
        except IndexError:
            head = s.repo.create_head(branch_name, b)

        head.set_tracking_branch(b)
        # Omits ISO
        s.repo.git.checkout("main", "--", "ADB", "ESTAT", "IAMC", "TDCI")


# Command-line interface


@click.group("store")
@click.pass_context
def main(context) -> None:
    """Manipulate local SDMX (meta)data."""


@main.command()
@click.pass_obj
def clone(context):
    """Clone the registry.

    The registry is cloned into the directory specified by the `tdc_registry_local`
    config value. See `tdc config --help`.
    """
    from transport_data import STORE

    STORE.clone()


@main.command("list")
@click.option("--maintainer", "maintainer_id")
@click.pass_context
def list_cmd(context, maintainer_id):
    """List contents of the store."""
    from transport_data import STORE

    for urn in sorted(STORE.list(maintainer=maintainer_id)):
        print(sdmx.urn.shorten(urn))


@main.command()
@click.argument("partial_urn", metavar="URN")
@click.pass_context
def show(context, partial_urn):
    """Display an SDMX object by URN.

    The URN should be partial, starting with the object class, for instance
    "Codelist=AGENCY:ID(1.2.3)".

    .. todo:: Use :class:`.Report` instances where possible.
    """
    from transport_data import STORE

    obj = STORE.get(partial_urn)

    print(repr(obj))

    if isinstance(obj, m.ItemScheme):
        # Display members of the ItemScheme
        for i, (_, item) in enumerate(obj.items.items()):
            print(f"{i:>3} {repr(item)}")



"""CLI for :mod:`.org`.

Use the :program:`--help` command-line option to see help for individual commands
defined in this module.

.. runblock:: console

   $ tdc org --help

"""

import pathlib

import click


@click.group("org")
def main():
    """TDCI itself."""


@main.command()
def refresh():
    """Update the TDCI metadata."""
    from transport_data import STORE

    from . import get_agencyscheme

    STORE.set(get_agencyscheme())


@main.command("read")
@click.argument(
    "path", type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path)
)
def read(path: "pathlib.Path"):
    """Read and summarize metadata."""
    from .metadata import report
    from .metadata.spreadsheet import read_workbook

    mds, _ = read_workbook(path.resolve())

    print(report.MetadataSet0Plain(mds).render())


@main.command
@click.option("--ref-area", required=True)
@click.argument(
    "path_in", type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path)
)
@click.argument(
    "path_out", type=click.Path(dir_okay=False, path_type=pathlib.Path), required=False
)
def summarize(path_in: "pathlib.Path", path_out: "pathlib.Path | None", ref_area):
    """Generate HTML metadata summary.

    If a single value is given for --ref-area (e.g. --ref-area=CA), a summary is
    generated of the (meta)data pertaining to that country/area. If multiple values are
    given (e.g. --ref-area=AF,ZW), a summary table is generated.
    """
    from .metadata import report
    from .metadata.spreadsheet import read_workbook

    mds, _ = read_workbook(path_in.resolve())

    ref_areas = ref_area.split(",")
    if 1 == len(ref_areas):
        # Report for a single REF_AREA
        if path_out is None:
            path_out = pathlib.Path.cwd().joinpath(f"{ref_areas[0]}.{{html,odt}}")
            print(f"Write to {path_out}")
        report.MetadataSet1HTML(mds, ref_area=ref_areas[0]).write_file(
            path_out.with_suffix(".html"), encoding="utf-8"
        )
        report.MetadataSet1ODT(mds, ref_area=ref_areas[0]).write_file(
            path_out.with_suffix(".odt")
        )
    elif 1 < len(ref_areas):
        # Report for multiple REF_AREA
        if path_out is None:
            path_out = pathlib.Path.cwd().joinpath("all.{html,odt}")
            print(f"Write to {path_out}")
        report.MetadataSet0ODT(mds).write_file(path_out.with_suffix(".odt"))
        report.MetadataSet2HTML(mds, ref_area=ref_areas).write_file(
            path_out.with_suffix(".html"), encoding="utf-8"
        )


@main.command("tuewas")
@click.argument(
    "path_in", type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path)
)
def _tuewas_all(path_in):
    """Generate all outputs for TUEWAS.

    PATH_IN is the path to the .xlsx file containing collected metadata.
    """
    from zipfile import ZipFile

    from transport_data.util.libreoffice import HAS_LIBREOFFICE

    from .metadata import merge_ato, report
    from .metadata.spreadsheet import read_workbook

    ref_areas = "CN ID IN PH TH VN".split()

    # Read collected metadata for the project
    mds, _ = read_workbook(path_in.resolve())

    # Merge metadata available through ATO
    merge_ato(mds)

    dir_out = pathlib.Path.cwd().joinpath("output")
    path_out = []

    def _maybe_pdf():
        if HAS_LIBREOFFICE:
            # Also mark the auto-converted PDF to be added to the ZIP archive
            path_out.append(path_out[-1].with_suffix(".pdf"))
            print(f"Wrote {path_out[-2:]}")
        else:
            print(f"Wrote {path_out[-1]}")

    for ref_area in ref_areas:
        path_out.append(dir_out.joinpath(ref_area, "Summary.odt"))
        path_out[-1].parent.mkdir(parents=True, exist_ok=True)
        report.MetadataSet1ODT(mds, ref_area=ref_area).write_file(path_out[-1])
        _maybe_pdf()

    path_out.append(dir_out.joinpath("Metadata summary.odt"))
    report.MetadataSet0ODT(mds).write_file(path_out[-1])
    _maybe_pdf()

    path_out.append(dir_out.joinpath("Metadata summary table.html"))
    report.MetadataSet2HTML(mds, ref_area=ref_areas).write_file(
        path_out[-1], encoding="utf-8"
    )
    print(f"Wrote {path_out[-1]}")

    path_zip = dir_out.joinpath("all.zip")
    with ZipFile(path_zip, mode="w") as zf:
        for p in path_out:
            zf.write(p, str(p.relative_to(dir_out)))

    print(f"Wrote {path_zip}")


@main.command("template")
def template():
    """Generate the metadata template."""
    from .metadata.spreadsheet import make_workbook

    make_workbook()



"""CLI for :mod:`.ato`.

.. runblock:: console

   $ tdc ato --help

.. runblock:: console

   $ tdc ato fetch --help

.. runblock:: console

   $ tdc ato convert --help
"""

import click

from . import FILES, convert, fetch


@click.group("ato")
def main():
    """Asian Transport Observatory (ATO) provider."""


@main.command("fetch")
@click.argument("part", nargs=-1)
@click.option("--go", is_flag=True, help="Actually fetch.")
@click.option("--all", "all_", is_flag=True, help="Fetch all files.")
@click.option(
    "--from-zenodo", "from_zenodo", is_flag=True, help="Fetch from Zenodo mirror."
)
def fetch_cmd(part, all_, go: bool, from_zenodo: bool) -> None:
    """Fetch original data files."""
    if not len(part):
        if not all_:
            print(f"Supply --all or 1+ of: {' '.join(FILES)}")
            return

        part = list(FILES.keys())

    fetch(*part, from_zenodo=from_zenodo, dry_run=not go)


@main.command("convert")
@click.argument("part", nargs=-1)
@click.option("--all", "all_", is_flag=True, help="Convert all parts.")
def convert_cmd(part, all_):  # pragma: no cover
    """Convert to SDMX data and structures."""
    if not len(part):
        if not all_:
            print(f"Supply --all or 1+ of: {' '.join(FILES)}")
            return

        part = list(FILES.keys())

    for p in part:
        if p == "POL":
            print(f"Skip PART = {p}; not currently supported")
            continue

        convert(p)



"""Information about the TDCI *per se*."""

from datetime import date
from itertools import chain
from typing import TYPE_CHECKING

import sdmx.model.v21 as m

from transport_data.util.pluggy import hookimpl, pm

if TYPE_CHECKING:
    import sdmx.model.v21


@hookimpl
def get_agencies() -> "sdmx.model.v21.Agency":
    """Return agencies and organizations including and subsidiary to TDCI itself."""
    # Agency
    a1 = m.Agency(
        id="TDCI",
        name="Transport Data Commons Initiative",
        description="See https://transport-data.org",
    )
    c1 = m.Contact(
        responsibility="Organization team",
        email=["kirsten.orschulok@giz.de", "verena.knoell@giz.de"],
        uri=["https://transport-data.org"],
    )
    a1.contact.append(c1)

    a2 = m.Agency(id="EPT", name="Expert Prototype Team")

    c2 = m.Contact(
        name="Paul Natsuo Kishimoto",
        email=["mail@paul.kishimoto.name"],
        responsibility="Admin for https://github.com/transport-data",
    )
    c3 = m.Contact(name="Marie Colson", email=["marie.colson@ifeu.de"])
    c4 = m.Contact(name="Pierpaolo Cazzola", email=["pierpaolo.cazzola@gmail.com"])
    c5 = m.Contact(name="James Dixon", email=["james.dixon@ouce.ox.ac.uk"])
    c6 = m.Contact(name="Dominic Sheldon", email=["dominic.sheldon@ricardo.com"])
    c7 = m.Contact(name="Alex Blackburn", email=["blackburna@un.org"])
    c8 = m.Contact(name="François Cuenot", email=["francois.cuenot@un.org"])
    a2.contact.extend([c2, c3, c4, c5, c6, c7, c8])

    a1.append_child(a2)

    return a1, a2


@hookimpl
def provides():
    return ("AgencyScheme=TDCI:TDCI",)


def get_agencyscheme() -> "sdmx.model.v21.AgencyScheme":
    """Generate an AgencyScheme including some TDCI data providers."""
    as_ = m.AgencyScheme(
        id="TDCI",
        # NameableArtefact
        name="Transport Data Commons Initiative partners and participants",
        # VersionableArtefact
        version="2025.12.27",
        valid_from=date.today().isoformat(),
        # MaintainableArtefact
        maintainer=None,
    )

    for agency in chain(*pm.hook.get_agencies()):
        as_.append(agency)

    # TDCI itself is the maintainer
    as_.maintainer = as_["TDCI"]

    return as_


def refresh():
    """Refresh the registry with structures from this module."""
    from transport_data import STORE

    as_ = get_agencyscheme()
    STORE.set(as_)



"""Plug-in hooks to be implemented by submodules and other packages."""

from typing import TYPE_CHECKING, Iterable

import pluggy

if TYPE_CHECKING:
    import sdmx.model.v21

hookspec = pluggy.HookspecMarker("transport_data")


@hookspec
def get_agencies() -> Iterable["sdmx.model.v21.Agency"]:
    """Return :class:`sdmx.model.common.Agency` identifying (meta)data provider(s).

    An implementation **must** return an iterable of 0 or more Agency instances.
    """
    raise NotImplementedError


@hookspec
def provides() -> Iterable[str]:
    """Return the URNs of SDMX artefacts available from a module."""
    raise NotImplementedError



"""Handle data and structure for IAMC-like formats.

.. todo:: Add a function to generate distinct DSDs for each data flow in a data set.
.. todo:: Add function(s) to reshape IAMC-like data.
"""

import logging

import pandas as pd
import sdmx.model.v21 as m
from sdmx.message import StructureMessage

from transport_data.util.pluggy import hookimpl

log = logging.getLogger(__name__)


@hookimpl
def get_agencies():
    a = m.Agency(
        id="IAMC",
        name="Integrated Assessment Modeling Consortium",
        contact=[m.Contact(uri=["https://iamconsortium.org"])],
    )
    return (a,)


@hookimpl
def provides():
    return ("ConceptScheme=TDCI:CS_IAMC",)


def common_structures():
    """Return common metadata for IAMC-like data and structures.

    Returns
    -------
    ConceptScheme
        with id "IAMC", containing the concepts for the IAMC dimensions and attribute.
    """
    cs = m.ConceptScheme(
        id="IAMC", name="Concepts in the IAMC data model", maintainer=get_agencies()[0]
    )

    cs.extend(
        [
            m.Concept(id="MODEL", name="Model name"),
            m.Concept(id="SCENARIO", name="Model configuration or parametrization"),
            m.Concept(id="REGION", name="Geographical area"),
            m.Concept(id="VARIABLE", name="Measure and additional dimensions"),
            m.Concept(id="UNIT", name="Unit of measurement"),
            m.Concept(id="YEAR", name="Time period"),
        ]
    )

    return cs


def structures_for_data(
    data: pd.DataFrame,
    base_id: str = "GENERATED",
    maintainer: m.Agency | None = None,
) -> StructureMessage:
    """Return IAMC-like data structures describing `data`.

    Parameters
    ----------
    data : pandas.DataFrame
        Data in IAMC tabular format.

        - Either long (with a "YEAR" column) or wide (no "YEAR" column but one or more
          columns with :class:`int` codes for the "YEAR" dimension).
        - Column names in any case.
          Upper-cased column names must appear in the IAMC concept scheme
          (:func:`get_iamc_structures`).
    maintainer : .Agency, optional
        Maintainer to be associated with generated :mod:`.MaintainableArtefact`.

    Returns
    -------
    StructureMessage
        including:

        - The "IAMC" ConceptScheme from :func:`get_iamc_structures`.
        - A data structure definition with ID `base_id`.
        - One code list for each dimension.
        - A "MEASURE" concept scheme including the unique measures: that is, parts
          appearing before the first pipe ("|") separator in the "VARIABLE" column of
          `data`.
        - For each measure: 0 or more code lists, each containing codes for a single
          dimension in the parts of "VARIABLE" column values beyond the first pipe ("|")
          separator.
    """
    # Generic IAMC ConceptScheme
    iamc_cs = common_structures()

    # Default maintainer
    maintainer = maintainer or m.Agency(id="TEST")

    # Empty structures
    sm = StructureMessage()
    dsd = m.DataStructureDefinition(id=base_id, maintainer=maintainer)

    # Identify dimension IDs from column names
    data_format = "long"
    for name in data.columns:
        try:
            # Identify `name` as one of the standard IAMC concepts
            dim_concept = iamc_cs[name.upper()]
            dsd.dimensions.append(m.Dimension(id=name, concept_identity=dim_concept))
            continue
        except KeyError:
            pass
        try:
            # Identify `name` as an integer year label
            int(name)
            # Succeeded; add the YEAR dimension if not already added
            if "YEAR" not in dsd.dimensions:
                data_format = "wide"
                dsd.dimensions.append(
                    m.Dimension(id="YEAR", concept_identity=iamc_cs["YEAR"])
                )
        except ValueError:
            raise

    dsd.description = f"The original data are in {data_format!r} format."
    sm.add(dsd)

    # Create code lists for other dimensions
    for dim in filter(lambda d: d.id.upper() not in "YEAR VARIABLE", dsd.dimensions):
        sm.add(
            cl_for_data(data[dim.id], id=dim.concept_identity.id, maintainer=maintainer)
        )

    # Special handling for "VARIABLE"
    for obj in structures_for_variable(data["variable"], maintainer=maintainer):
        sm.add(obj)

    sm.add(iamc_cs)

    return sm


def structures_for_variable(data: pd.Series, **ma_kwargs) -> list:
    """Make structures for IAMC-like "VARIABLE" `data`.

    Parameters
    ----------
    ma_kwargs :
        Keyword arguments for :class:`.MaintainableArtefact`.

    Returns
    -------
    list
        A sequence of SDMX structures.
    """
    # Split the data on the first "|"
    parts = data.drop_duplicates().str.split("|", expand=True)

    # A Concept scheme for the measures appearing in `data`
    cs = m.ConceptScheme(id="MEASURE", **ma_kwargs)

    # Structures to be returned
    structures = [cs]

    # Identify and group by the measure in the first part
    for name, group_parts in parts.groupby(0):
        # Add the measure to the concept scheme
        measure = m.Concept(id=str(name).upper().replace(" ", "_"), name=name)
        cs.append(measure)

        # Make a DSD and code lists from the remaining parts
        # TODO also include the general (model, scenario, etc.) dimensions
        structures.extend(
            structures_for_measure(measure, group_parts.iloc[:, 1:], **ma_kwargs)
        )

    log.info(
        f"Identified {len(cs)} measures from {len(parts)} distinct 'variable' values"
    )

    return structures


def structures_for_measure(
    measure: m.Concept, parts: pd.DataFrame, **ma_kwargs
) -> list:
    """Create a DSD and code lists for a particular `measure` from variable `parts`.

    Parameters
    ----------
    ma_kwargs :
        Keyword arguments for :class:`.MaintainableArtefact`.
    """
    dsd = m.DataStructureDefinition(id=measure.id, **ma_kwargs)

    # Structures to be returned
    structures = [dsd]

    # Drop any columns that are empty
    parts = parts.dropna(axis=1, how="all")

    # Check whether labels are balanced
    some_empty = parts.isna().any(axis=0)
    if some_empty.any():
        # commented: log information about unbalanced numbers of dimensions
        N_dim = parts.shape[1]
        N_empty = some_empty.sum()

        log.info(f"Measure {measure.id!r}")
        log.info(
            f"Variables contain uneven dimension labels: {N_dim - N_empty}–{N_dim}"
        )
        pass

    # Iterate over remaining dimensions
    for i, s in parts.items():
        # Numerical ID for this dimension
        dim_id = f"DIM_{i}"

        # Generate a code list
        cl = cl_for_data(s.dropna(), id=f"{measure.id}_{dim_id}", **ma_kwargs)

        # Add a special value for missing labels "_"
        if some_empty[i]:  # type: ignore [call-overload]
            cl.append(m.Code(id="_", name="No label/missing"))
        structures.append(cl)

        # Generate a dimension that references this code list and measure
        dsd.dimensions.append(
            m.Dimension(
                id=dim_id,
                concept_identity=measure,
                local_representation=m.Representation(enumerated=cl),
            )
        )

    return structures


def cl_for_data(data: pd.Series, id: str, **ma_kwargs) -> m.Codelist:
    """Make a codelist for the concept `id`, given `data`.

    Parameters
    ----------
    ma_kwargs :
        Keyword arguments for :class:`.MaintainableArtefact`.
    """
    cl = m.Codelist(id=id, **ma_kwargs)

    for value in sorted(data.unique()):
        cl.append(m.Code(id=value))

    return cl


def variable_cl_for_dsd(
    dsd: m.BaseDataStructureDefinition, codelist: m.Codelist | None = None
) -> m.Codelist:
    """Generate an SDMX codelist with IAMC "VARIABLE" codes corresponding to `dsd`.

    The dimensions of `dsd` are each enumerated by an associated codelist.
    :meth:`DataStructureDefinition.iter_keys
    <sdmx.model.common.BaseDataStructureDefinition.iter_keys>` generates one key for
    each member of the Cartesian product of these sets.

    :func:`.variable_cl_for_dsd` collapses each of these into a code whose ID is a
    constructed IAMC "VARIABLE" name like "Primary Measure|Foo|Bar|Baz". In these:

    - The part before the first pipe ("Primary Measure|") is the name of the concept
      identity for the primary measure of `dsd`.
    - The remaining parts ("Foo|Bar|Baz") are the names of the codes for each dimension
      of `dsd`.

    If the code lists that enumerate `dsd` contain codes with the special ID "_T", then
    :func:`.variable_cl_for_dsd` also generates corresponding IAMC-style names for
    aggregates. In the above example, if the third dimension contains a "_T"

    Every code also has annotations with the following IDs and text:

    ``iamc-full-dsd``
       The URN of `dsd`. This allows an unambiguous association to the
       full-dimensionality data structure definition.
    ``iamc-full-key``
       The :func:`repr` of a :class:`dict` corresponding to a full resolution key,
       mapping dimension ID to code ID; for example "{'dim_0': 'FOO', 'dim_1': 'BAR',
       'dim_2': 'BAZ'}". This allows to restore or recover a valid key within `dsd`,
       given the IAMC "VARIABLE".

    If `codelist` is supplied, the new codes are appended.

    Parameters
    ----------
    dsd :
        Existing data structure definition.
    codelist : Codelist, optional
        Existing code list.

    Returns
    -------
    Codelist
        The same object as `codelist`, if supplied, else a new code list with id
        "VARIABLE".
    """
    cl = codelist or m.Codelist(id="VARIABLE")

    for key in dsd.iter_keys():
        # Parts of the variable ID; full key as (str -> str)
        var_parts, full_key = [str(dsd.measures[0].concept_identity.name)], {}

        # Iterate over KeyValues
        for dim, kv in key.values.items():
            if kv.value.id == "_T":
                # Total: pass through
                var_parts.append("_T")
            else:
                # Use the (human-readable) name
                var_parts.append(str(kv.value.name))
            full_key[kv.value_for.id] = kv.value.id

        # - Join var_parts with the IAMC "|" character.
        # - Remove "|_T" to yield IAMC-style names for aggregates.
        variable = "|".join(var_parts).replace("|_T", "")

        # Create a Code
        # - ID is the variable name.
        # - Annotate with iamc-full-dsd = URN of the full DSD.
        # - Annotate with iamc-full-key = representation of the full key.
        cl.append(
            m.Code(
                id=variable,
                annotations=[
                    m.Annotation(id="iamc-full-dsd", text=dsd.urn),
                    m.Annotation(id="iamc-full-key", text=repr(full_key)),
                ],
            )
        )

    return cl



import pytest

from transport_data.iso import generate_all


@pytest.fixture(scope="session")
def all_generated():
    """All ISO code lists generated in the test data directory."""
    generate_all()


def test_generate_all(all_generated):
    from transport_data import STORE

    urns = STORE.list(maintainer="ISO")

    # At least 15 code lists were generated
    assert 15 <= len(urns)


@pytest.mark.parametrize(
    "standard_number, id_field, N_exp, entries, N_tr",
    (
        ("639-3", "alpha_2", 184, ("aa", "zu"), 57),
        ("639-3", "alpha_3", 7910, ("aaa", "zzj"), 16),
        #
        ("639-5", "alpha_3", 115, ("aav", "znd"), 34),
        #
        ("3166-1", "alpha_2", 249, ("AW", "ZE"), 58),
        ("3166-1", "alpha_3", 249, ("ABW", "ZWE"), 58),
        ("3166-1", "numeric", 249, ("533", "716"), 58),
        #
        ("3166-2", "code", 5046, ("AD-02", "ZW-MW"), 16),
        #
        ("3166-3", "alpha_2", 30, ("AI", "ZR"), 80),
        ("3166-3", "alpha_3", 31, ("AFI", "ZAR"), 80),
        ("3166-3", "alpha_4", 31, ("AIDJ", "ZRCD"), 80),
        ("3166-3", "numeric", 25, ("262", "180"), 80),
        #
        ("4217", "alpha_3", 181, ("AED", "ZWL"), 55),
        ("4217", "numeric", 181, ("784", "932"), 55),
        #
        ("15924", "alpha_4", 182, ("Adlm", "Zzzz"), 25),
        ("15924", "numeric", 182, ("166", "999"), 25),
    ),
)
def test_generate_codelist(
    all_generated,
    # Function arguments
    standard_number: str,
    id_field: str,
    # Expected values
    N_exp: int,  # Total number of items
    entries: tuple[str, ...],  # IDs of some items
    N_tr: int,  # Number of localizations for the name of the first of `entries`
) -> None:
    from transport_data import STORE

    # The code list is available with the expected URL
    cl = STORE.get(f"Codelist=ISO:{standard_number}_{id_field}")

    # Result has the expected number of codes
    assert N_exp == len(cl)

    # The name of the first of `entries` is localized in `N_tr` languages
    assert N_tr == len(cl[entries[0]].name.localizations)



import json
import os
import platform
import re
import zipfile
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING, cast

import click.testing
import pytest
import responses
import sdmx.message
from sdmx.model import common, v21

import transport_data
from transport_data.config import Config
from transport_data.store import UnionStore

if TYPE_CHECKING:
    from importlib.resources.abc import Traversable

    import dsss.store

    from transport_data.util.sdmx import MAKeywords

#: List of CKAN UUIDs for objects used for testing.
CKAN_UUID = {
    "DEV org test": "538d7317-fa7b-4890-a945-ae4d30021f38",
    "PROD org test": "8bc58a55-fd85-4667-b23e-a5dba18cdb9a",
}

#: :any:`True` if tests are being run on GitHub Actions.
GITHUB_ACTIONS: bool = "GITHUB_ACTIONS" in os.environ

#: Marks to be reused throughout the test suite. Do not reuse keys in the mapping.
MARK = {
    0: pytest.mark.xfail(
        condition=GITHUB_ACTIONS and platform.system() == "Windows",
        raises=zipfile.BadZipFile,
        reason="'Truncated file header' on GHA runner",
    ),
    "#52": pytest.mark.xfail(
        raises=AssertionError,
        reason="Upstream files removed; https://github.com/transport-data/tools/issues/52",
    ),
}


class CliRunner(click.testing.CliRunner):
    def invoke(self, *args, **kwargs):
        import transport_data.cli

        return super().invoke(transport_data.cli.main, *args, **kwargs)


def ember_dfd(store: "dsss.store.Store") -> "v21.Dataflow":
    """Construct a Dataflow used in testing :func:`.read_csv` and related code.

    .. todo:: Instead ensure these artefacts are available from :data:`.STORE`.
    """

    # Construct a DSD corresponding to the data
    ma_kw: "MAKeywords" = dict(
        id="EMBER_001", version="1.0.0", maintainer=common.Agency(id="TDCI")
    )
    dsd = v21.DataStructureDefinition(**ma_kw)
    dsd.dimensions.append(v21.Dimension(id="COUNTRY"))
    dsd.dimensions.append(v21.Dimension(id="YEAR"))
    dsd.measures.append(v21.PrimaryMeasure(id="OBS_VALUE"))
    dsd.attributes.append(common.DataAttribute(id="COMMENT"))

    # Construct a DFD
    dfd = v21.DataflowDefinition(**ma_kw, structure=dsd)

    store.set(dfd)
    store.set(dsd)

    return dfd


@pytest.fixture(scope="session")
def mock_zenodo_api(
    test_data_path: "Traversable",
) -> Iterator["responses.RequestsMock"]:
    """Mock Zenodo API responses to avoid hitting rate limits."""
    from transport_data.util.responses import RepeatRegistry, RequestsMock

    # Use a registry that repeats the same response for a given URL
    mock = RequestsMock(assert_all_requests_are_fired=False, registry=RepeatRegistry)

    # Pass through requests to certain other domains
    mock.add_passthru("https://asiantransportobservatory.org")
    mock.add_passthru("https://doi.org")
    # Pass through requests for actual files
    mock.add_passthru(
        re.compile(r"^https://zenodo\.org/api/records/\d+/files/.*/content$")
    )

    # Mock API responses with record information
    base = "https://zenodo.org/api"
    for doi in "14913730", "15232577":
        with open(str(test_data_path.joinpath(f"zenodo-{doi}.json"))) as f:
            mock.add(method="GET", url=f"{base}/records/{doi}", json=json.load(f))

    yield mock


@pytest.fixture(scope="session")
def sdmx_structures(tmp_store) -> sdmx.message.StructureMessage:
    """SDMX structures for use in tests."""
    sm = sdmx.message.StructureMessage()

    ma_attrib: "MAKeywords" = dict(maintainer=common.Agency(id="TEST"), version="1.0.0")

    cs = common.ConceptScheme(id="TEST", **ma_attrib)  # type: ignore [misc]
    cs.append(common.Concept(id="MASS", name="Mass of fruit"))
    cs.append(common.Concept(id="PICKED", name="Number of fruits picked"))
    cs.append(common.Concept(id="COLOUR", name="Colour of fruit"))
    cs.append(common.Concept(id="FRUIT", name="Type of fruit"))

    cl: "common.Codelist" = common.Codelist(id="COLOUR", **ma_attrib)  # type: ignore [misc]
    cl.extend(
        [common.Code(id=c, name=c.title()) for c in "GREEN ORANGE RED YELLOW".split()]
    )
    cl.append(common.Code(id="_T", name="Total"))
    sm.add(cl)

    cl = common.Codelist(id="FRUIT", **ma_attrib)  # type: ignore [misc]
    cl.extend(
        [common.Code(id=c, name=c.title()) for c in "APPLE BANANA GRAPE LEMON".split()]
    )
    cl.append(common.Code(id="_T", name="Total"))
    sm.add(cl)

    for id_, dims in (
        ("MASS", ("COLOUR", "FRUIT")),
        ("PICKED", ("FRUIT", "COLOUR")),
    ):
        dsd = v21.DataStructureDefinition(id=id_, **ma_attrib)  # type: ignore [misc]
        dsd.urn = sdmx.urn.make(dsd)
        dsd.measures.append(v21.PrimaryMeasure(id=id_, concept_identity=cs[id_]))
        dsd.dimensions.extend(
            common.Dimension(
                id=d,
                concept_identity=cs[d],
                local_representation=common.Representation(enumerated=sm.codelist[d]),
            )
            for d in dims
        )
        sm.add(dsd)

    tmp_store.update_from(sm)

    return sm


@pytest.fixture
def tdc_cli():
    """A :class:`.CliRunner` object that invokes the :program:`tdc` CLI."""
    yield CliRunner()


@pytest.fixture(scope="session")
def test_data_path() -> Iterator["Traversable"]:
    """Path containing test data."""
    from importlib.resources import files

    # TODO When Python 3.11 is the minimum supported, use separate "data", "tests" args
    yield files("transport_data").joinpath("data/tests")


@pytest.fixture(scope="session")
def tmp_config(tmp_path_factory) -> Generator[Config, None, None]:
    """A :class:`.Config` instance pointing to a temporary directory."""
    from platformdirs import user_data_path

    base = tmp_path_factory.mktemp("transport-data")
    result = Config(
        cache_path=base.joinpath("cache"),
        config_path=base.joinpath("config.json"),
        data_path=base.joinpath("data"),
        # Default value in .config.Config.data_path → clone from another local directory
        registry_remote_url=str(user_data_path("transport-data").joinpath("registry")),
    )

    with pytest.MonkeyPatch().context() as mp:
        mp.setattr(transport_data, "CONFIG", result)
        yield result


@pytest.fixture(scope="session")
def tmp_store(tmp_config) -> Generator[UnionStore, None, None]:
    """A :class`.UnionStore` in a temporary directory per :func:`.tmp_config`."""
    result = UnionStore(tmp_config)

    # Initialize an empty Git repo
    cast("dsss.store.GitStore", result.store["registry"]).path.mkdir(exist_ok=True)

    with pytest.MonkeyPatch().context() as mp:
        mp.setattr(transport_data, "STORE", result)
        yield result



import logging
import sys

from .config import Config
from .store import UnionStore
from .util.pluggy import register_internal

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler(sys.stdout))

#: Global configuration.
CONFIG = Config.read()

#: Global access to data storage.
STORE = UnionStore(CONFIG)

# Register plugin hooks
register_internal(
    "ato",
    "iamc",
    "ipcc",
    "iso",
    "itdp",
    "jrc",
    "oica",
    "org",
    "other",
)



"""CLI for :mod:`.jrc`.

.. runblock:: console

   $ tdc jrc --help

.. runblock:: console

   $ tdc jrc fetch --help

.. runblock:: console

   $ tdc jrc convert --help
"""

import click

from . import GEO, convert, fetch


@click.group("jrc")
def main():
    """EU Joint Research Center (JRC) provider."""


@main.command("fetch")
@click.argument("geo", nargs=-1)
@click.option("--go", is_flag=True, help="Actually fetch.")
@click.option("--all", "all_", is_flag=True, help="Fetch all files.")
def fetch_cmd(geo, all_, go):
    """Fetch original data files."""
    if not len(geo):
        if not all_:
            print(f"Supply --all or 1+ of {' '.join(sorted(GEO))}")
            return

        geo = GEO

    fetch(*geo, dry_run=not go)


@main.command("convert")
@click.argument("geo", nargs=-1)
@click.option("--all", "all_", is_flag=True, help="Fetch all files.")
def convert_cmd(geo, all_):
    """Convert to SDMX data and structures."""
    if not len(geo):
        if not all_:
            print(f"Supply --all or 1+ of {' '.join(sorted(GEO))}")
            return

        geo = GEO

    try:
        for g in geo:
            convert(g)
    except AssertionError:
        raise click.Abort



from dataclasses import dataclass

import pytest
import sdmx
from sdmx.model import v21

from transport_data.testing import ember_dfd
from transport_data.util.sdmx import fields_to_mda, read_csv


def test_fields_to_mda() -> None:
    mdsd = v21.MetadataStructureDefinition()
    rs = mdsd.report_structure["ALL"] = v21.ReportStructure(id="ALL")

    assert 0 == len(rs)

    @dataclass
    class MDSExample:
        #: Foo
        #:
        #: Description of Foo.
        foo: str

        bar: int

    fields_to_mda(MDSExample, rs)

    # Report structure has 1 metadata attribute for each of the fields in MDSExample
    assert 2 == len(rs)

    # Metadata attributes have expected IDs
    assert "foo" == rs.components[0].id == rs.components[0].concept_identity.id
    assert "bar" == rs.components[1].id == rs.components[1].concept_identity.id

    # Python type annotations are stored as 'data-type' SDMX annotations
    assert "<class 'str'>" == rs.components[0].eval_annotation(id="data-type")
    assert "<class 'int'>" == rs.components[1].eval_annotation(id="data-type")

    # Python docstrings are stored as name and description on the ConceptIdentity
    assert "Foo" == str(rs.components[0].concept_identity.name)
    assert "Description of Foo." == str(rs.components[0].concept_identity.description)
    assert 0 == len(rs.components[1].concept_identity.name.localizations)
    assert 0 == len(rs.components[1].concept_identity.description.localizations)


@pytest.mark.parametrize(
    "filename, adapt",
    (
        # SDMX-CSV
        ("read-csv-0.csv", {}),
        # Simplified CSV with CSVAdapter arguments
        (
            "read-csv-1.csv",
            dict(
                structure="dataflow",
                structure_id="TDCI:EMBER_001(1.0.0)",
                action="I",
            ),
        ),
    ),
)
def test_read_csv(test_data_path, tmp_store, filename, adapt) -> None:
    dfd = ember_dfd(tmp_store)

    # CSV data are read without error
    dm = read_csv(test_data_path.joinpath(filename), dfd, adapt)

    # A single data set is returned
    assert 1 == len(dm.data)

    # Data have the expected number of observations
    ds = dm.data[0]
    assert 120 == len(ds)

    # Data are structured by the given DFD and DSD
    assert set(ds.obs[0].dimension.values.keys()) == set(
        d.id for d in dfd.structure.dimensions
    )

    # Data can be converted to pandas
    df = sdmx.to_pandas(ds, attributes="o")

    # Observation values and attributes are accessible and have expected type and values
    assert 599.0 == df.loc[("China", "2021"), "value"]
    assert "FOO" == df.loc[("China", "2021"), "COMMENT"]



"""Handle TDC-structured metadata."""

import itertools
import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache, partial
from itertools import count
from typing import Any, Callable, Hashable, Iterable, cast

from pycountry import countries
from sdmx.model import common, v21

from transport_data.util.sdmx import fields_to_mda

log = logging.getLogger(__name__)


@dataclass
class TDCALLReportStructure:
    """Concepts and metadata attributes in the TDC metadata structure."""

    #: Data flow ID
    #:
    #: A unique identifier for the data flow (=data source, data set, etc.).
    #:
    #: We suggest to use IDs like ‘VN001’, where ‘VN’ is the ISO 3166 alpha-2 country
    #: code, and ‘001’ is a unique number. The value MUST match the name of the sheet
    #: in which it appears.
    DATAFLOW: str

    #: Data provider
    #:
    #: Organization or individual that provides the data and any related metadata.
    #:
    #: This can be as general (“IEA”) or specific (organization unit/department,
    #: specific person responsible, contact details, etc.) as appropriate.
    DATA_PROVIDER: str

    #: URL or web address
    #:
    #: Location on the Internet with further information about the data flow.
    URL: str

    #: Measure (‘indicator’)
    #:
    #: Statistical concept for which data are provided in the data flow.
    #:
    #: If the data flow contains data for multiple measures, give each one separated by
    #: semicolons. Example: “Number of cars; passengers per vehicle”.
    #:
    #: This SHOULD NOT duplicate the value for ‘UNIT_MEASURE’. Example: “Annual driving
    #: distance per vehicle”, not “Kilometres per vehicle”.
    MEASURE: str

    #: Unit of measure
    #:
    #: Unit in which the data values are expressed.
    #:
    #: If ‘MEASURE’ contains 2+ items separated by semicolons, give the respective units
    #: in the same way and order. If there are no units, write ‘dimensionless’, ‘1’, or
    #: similar.
    UNIT_MEASURE: str

    #: Dimensions
    #:
    #: Formally, the “statistical concept used in combination with other statistical
    #: concepts to identify a statistical series or individual observations.
    #:
    #: Record all dimensions of the data, either in a bulleted or numbered list, or
    #: separated by semicolons. In parentheses, give some indication of the scope
    #: and/or resolution of the data along each dimension. Most data have at least time
    #: and space dimensions.
    #:
    #: Example:
    #:
    #: - TIME_PERIOD (annual, 5 years up to 2021)
    #: - REF_AREA (whole country; VN only)
    #: - Vehicle type (12 different types: […])
    #: - Emissions species (CO2 and 4 others)"""
    DIMENSION: str

    #: Data description
    #:
    #: Any information about the data flow that does not fit in other attributes.
    #:
    #: Until or unless other metadata attributes are added to this metadata structure/
    #: template, this MAY include:
    #:
    #: - Any conditions on data access, e.g. publicly available, proprietary, fee or
    #:   subscription required, available on request, etc.
    #: - Frequency of data updates.
    #: - Any indication of quality, including third-party references that indicate data
    #:   quality.
    DATA_DESCR: str

    #: Methodology
    #:
    #: Any information about methods used by the data provider to collect, process, or
    #: prepare the data.
    METHOD: str

    #: Comment
    #:
    #: Any other information about the metadata values, for instance discrepancies or
    #: unclear or missing information.
    #:
    #: Precede comments with initials; append to existing comments to keep
    #: chronological order; and include a date (for example, “2024-07-24”) if helpful.
    COMMENT: str


def contains_data_for(mdr: "v21.MetadataReport", ref_area: str) -> bool:
    """Return :any:`True` if `mdr` contains data for `ref_area`.

    :any:`True` is returned if any of the following:

    1. The referenced data flow definition has an ID that starts with `ref_area`.
    2. The country's ISO 3166 alpha-2 code, alpha-3 code, official name, or common name
       appears in the value of the ``DATA_DESCR`` metadata attribute.


    Parameters
    ----------
    ref_area : str
        ISO 3166 alpha-2 code for a country. Passed to
        :meth:`pycountry.countries.lookup`.
    """
    if mdr.attaches_to.key_values["DATAFLOW"].obj.id.startswith(ref_area):  # type: ignore [union-attr]
        return True

    # Valid identifiers for `country`: its ISO 3166 alpha-[23] codes and names
    # NB In pycountry 23.12.11, "common_name" would fall back to "official_name" or
    #    "name" if not explicitly defined. In 24.6.1, this was reverted and
    #    AttributeError is raised.
    country = countries.lookup(ref_area)
    values: Iterable[str] = filter(
        None,
        (
            getattr(country, a, None)
            for a in "alpha_2 alpha_3 common_name name official_name".split()
        ),
    )
    # Pattern to match in DATA_DESCR
    pat = re.compile("(" + "|".join(values) + ")")

    for ra in mdr.metadata:
        assert hasattr(ra, "value")
        if ra.value_for.id == "DATA_DESCR" and pat.search(str(ra.value)):
            return True

    return False


def dfd_id(mdr: "v21.MetadataReport") -> str:
    """Return the ID of the dataflow targeted by `mdr`."""
    assert mdr.attaches_to is not None
    return cast(
        "v21.TargetIdentifiableObject", mdr.attaches_to.key_values["DATAFLOW"]
    ).obj.id


@lru_cache
def get_cs_common() -> "common.ConceptScheme":
    """Create a shared concept scheme for the concepts referenced by dimensions.

    Concepts in this scheme have an annotation ``tdc-aka``, which is a list of alternate
    IDs recognized for the concept.
    """
    from transport_data.org import get_agencyscheme

    as_ = get_agencyscheme()
    cs = common.ConceptScheme(id="CONCEPTS", maintainer=as_["TDCI"])

    cs.setdefault(
        id="CONFIDENTIALITY",
        annotations=[v21.Annotation(id="tdc-aka", text=repr(["CONFIDIENTALITY"]))],
    )
    cs.setdefault(
        id="FUEL_TYPE",
        annotations=[v21.Annotation(id="tdc-aka", text=repr(["Fuel type"]))],
    )
    cs.setdefault(
        id="GEO",
        annotations=[
            v21.Annotation(
                id="tdc-aka",
                text=repr(
                    ["Area", "Country", "Country code", "ECONOMY", "REF_AREA", "Region"]
                ),
            )
        ],
    )
    cs.setdefault(
        id="SERVICE",
        annotations=[v21.Annotation(id="tdc-aka", text=repr(["FREIGHT_PASSENGER"]))],
    )
    cs.setdefault(
        id="TIME_PERIOD",
        annotations=[v21.Annotation(id="tdc-aka", text=repr(["Time", "Year"]))],
    )

    return cs


@lru_cache
def get_msd() -> "v21.MetadataStructureDefinition":
    """Generate and return the TDC metadata structure definition."""
    from transport_data import STORE
    from transport_data.org import get_agencyscheme

    TDCI = get_agencyscheme()["TDCI"]

    cs = common.ConceptScheme(id="METADATA_CONCEPTS", maintainer=TDCI)
    msd = v21.MetadataStructureDefinition(id="SIMPLE", version="1", maintainer=TDCI)
    rs = msd.report_structure["ALL"] = v21.ReportStructure(id="ALL")

    fields_to_mda(TDCALLReportStructure, rs, cs)

    # NB Currently not supported by sdmx1; results in an empty XML collection
    STORE.set(msd)
    STORE.set(cs)

    return msd


def groupby(
    mds: "v21.MetadataSet", key=Callable[["v21.MetadataReport"], Hashable]
) -> dict[Hashable, list["v21.MetadataReport"]]:
    """Group metadata reports in `mds` according to a `key` function.

    Similar to :func:`itertools.groupby`.
    """
    result: dict[Hashable, list["v21.MetadataReport"]] = defaultdict(list)
    for k, g in itertools.groupby(mds.report, key):
        result[k].extend(g)
    return result


def _get(mdr: "v21.MetadataReport", mda_id: str) -> str | None:
    """Retrieve from `mdr` the reported value of the metadata attribute `mda_id`."""
    for mda in mdr.metadata:
        if mda.value_for is not None and mda.value_for.id == mda_id:
            assert hasattr(mda, "value")  # Exclude ReportedAttribute without value attr
            return str(mda.value) if mda.value else None
    # No match
    return None


# TODO Keep this dictionary as small as possible, by making appropriate changes in the
# input metadata
RENAME = {
    "IRF - International Road Federation": "International Road Federation (IRF)",
    "UIC": "International Union of Railways (UIC)",
}


def make_ra(mda_id: str, value: Any) -> "v21.OtherNonEnumeratedAttributeValue":
    """Generate a ReportedAttribute for `mda_id` with the given `value`."""
    mda = get_msd().report_structure["ALL"].get(mda_id)
    return v21.OtherNonEnumeratedAttributeValue(value=str(value), value_for=mda)


def make_tok(dfd: "common.BaseDataflow") -> "v21.TargetObjectKey":
    """Generate a :class:`.TargetObjectKey` that refers to `dfd`."""
    iot = v21.IdentifiableObjectTarget()
    return v21.TargetObjectKey(
        key_values={"DATAFLOW": v21.TargetIdentifiableObject(value_for=iot, obj=dfd)}
    )


def map_values_to_ids(mds: "v21.MetadataSet", mda_id: str) -> dict[str, set[str]]:
    """Return a mapping from unique reported attribute values to data flow IDs."""
    result = defaultdict(set)

    for r in mds.report:
        value = _get(r, mda_id) or "MISSING"

        # Merge or relabel certain values
        value = RENAME.get(value, value)

        result[value].add(_get(r, "DATAFLOW") or "MISSING")

    return result


def map_dims_to_ids(mds: "v21.MetadataSet") -> dict[str, set[str]]:
    """Return a mapping from unique concept IDs used for dimensions to data flow IDs."""
    result = defaultdict(set)

    for r in mds.report:
        dfd = r.attaches_to.key_values["DATAFLOW"].obj  # type: ignore [union-attr]
        for dim in dfd.structure.dimensions:
            key = f"{dim.id!r}"
            try:
                anno = dim.get_annotation(id="tdc-original-id")
                key += f" (as '{anno.text!s}')"
            except KeyError:
                pass

            result[key].add(dfd.id)

    return result


def unique_dfd_id(mdr: "v21.MetadataReport", existing: set[str]) -> str:
    """Generate a unique DSD ID for `mdr`."""
    template = f"{dfd_id(mdr)[:2]}{{:03d}}"
    for i in count():
        candidate = template.format(i)
        if candidate not in existing:
            existing.add(candidate)
            break
    return candidate


def merge_ato(mds: "v21.MetadataSet") -> None:
    """Extend `mds` with metadata reports for ATO/ATO data flows."""
    from transport_data import STORE
    from transport_data.ato import dataset_to_metadata_reports

    assert mds.structured_by

    N = len(mds.report)  # Initial number of reports
    all_dfd_ids = set(map(dfd_id, mds.report))  # Existing IDs of DFDs

    # Silence warnings from SDMX
    logging.getLogger("sdmx").setLevel(logging.ERROR)

    # Mapping from upstream URLs to metadata reports
    reports = defaultdict(list)

    # Iterate over existing data sets
    for ds in map(
        STORE.get,
        filter(
            lambda k: True,  # All results
            # lambda k: "-VEP-" in k,  # DEBUG Filter for a subset of data sets
            STORE.list(v21.DataSet, maintainer="ATO"),
        ),
    ):
        try:
            # Convert the attributes of `ds` into 1+ metadata report(s)
            for mdr in dataset_to_metadata_reports(ds, mds.structured_by):
                # Split URLs on multiple lines to a list
                for url in (_get(mdr, "URL") or "").split("\n"):
                    # Store a reference to each report in a collection for each URL
                    reports[url].append(mdr)
        except KeyError:
            # log.debug(f"Not in local store: {e.args[0]}")
            continue

    # Reports to attend to `mds` last
    sort_last = []

    # Iterate over URLs and lists of reports
    for url, report_group in reports.items():
        # Single report associated with this URL, or multiple reports but not with an
        # ID prefix like "CA_" that indicates a country-specific generated report
        if 1 == len(report_group) or not dfd_id(report_group[0])[2] == "_":
            mds.report.extend(report_group)  # Store directly
            continue

        # Identify the first report in the group
        mdr = report_group[0]

        # Identify the DATA_DESCR reported attribute
        for mda in mdr.metadata:
            if mda.value_for is not None and mda.value_for.id == "DATA_DESCR":
                break

        # Extend DATA_DESCR of `mdr` with values from other reports in `report_group`
        for other_value in filter(
            None, map(partial(_get, mda_id="DATA_DESCR"), report_group[1:])
        ):
            assert isinstance(mda, v21.OtherNonEnumeratedAttributeValue)
            mda.value = (mda.value or "") + "\n\n" + other_value

        # Replace the ID of `mdr` with a new, unique DFD ID including the country code
        assert mdr.attaches_to is not None
        cast(
            "v21.TargetIdentifiableObject", mdr.attaches_to.key_values["DATAFLOW"]
        ).obj.id = unique_dfd_id(mdr, all_dfd_ids)

        sort_last.append(mdr)

    mds.report.extend(sort_last)

    log.info(f"Appended {len(mds.report) - N} metadata reports for ATO data flows")

    # Restore log level
    logging.getLogger("sdmx").setLevel(logging.WARNING)



"""International Organization for Standardization (ISO)."""

import logging
from functools import partial
from typing import TYPE_CHECKING, Mapping

from sdmx.model import common, v21

from transport_data.util.pluggy import hookimpl
from transport_data.util.pycountry import LOCALIZABLE, get_database, load_translations

if TYPE_CHECKING:
    import gettext


log = logging.getLogger(__name__)


@hookimpl
def get_agencies():
    """Return the ``ISO`` :class:`~.sdmx.model.common.Agency`."""
    a = common.Agency(
        id="ISO",
        name="International Organization for Standardization",
        contact=[common.Contact(uri=["https://iso.org"])],
    )
    return (a,)


def generate_all() -> None:
    """Generate codelists for all the databases in :data:`.pycountry.DATABASE_INFO`."""
    from transport_data.util.pycountry import DATABASE_INFO

    for info in DATABASE_INFO:
        generate_codelists(standard_number=info.root_key)


def generate_codelists(standard_number: str) -> None:
    """Generate 1 or more :class:`.Codelist` with entries from ISO `standard_number`.

    Codes have:

    - IDs according to `id_field`.
    - An :attr:`~sdmx.model.common.IdentifiableArtefact.name` attribute localized to all
      the languages present in the upstream database.
    - Annotations for all other fields in the upstream database.

    Parameters
    ----------
    standard_number :
        ISO standard number, e.g. "3166-2" for ISO 3166-2.
    """
    from importlib.metadata import version

    import sdmx.urn

    from transport_data import STORE

    db, info = get_database(standard_number)

    # Load localizations for this standard number
    tr = partial(localize_all, translations=load_translations(f"iso{db.root_key}"))

    # Create 1 or more empty code lists: 1 for each possible id_field
    kw = dict(maintainer=get_agencies()[0], version=version("pycountry"))
    cl: dict[str, common.Codelist] = {
        f: common.Codelist(id=f"{db.root_key}_{f}", **kw) for f in info.id_fields
    }

    # Convert all entries in the database to SDMX Codes
    for record in db:
        # - Iterate over all fields in the record.
        # - For localizable fields, collect localizations of the field's value.
        # - Convert to Annotation objects. Not all of these will be used for each Code.
        anno = {
            f: v21.Annotation(id=f, text=tr(v) if f in LOCALIZABLE else {"zxx": v})
            for f, v in record._fields.items()
            if v is not None
        }

        # Add one code to cl["alpha_2"] with an alpha_2 ID, one to cl["alpha_3"] with an
        # alpha_3 ID, etc.
        for id_field in info.id_fields:
            try:
                id_ = anno[id_field].text.localized_default()
            except KeyError:
                # log.debug(f"No {id_field!r} field → omit:\n{record}")
                continue

            # Create a code
            c = common.Code(
                id=id_,
                name=anno["name"].text,
                annotations=[v for f, v in anno.items() if f not in {"name", id_field}],
            )

            # Append to the respective code list
            try:
                cl[id_field].append(c)
            except ValueError:
                log.info(
                    f"ID {id_!r} duplicates existing entry {cl[id_field][id_]!r} → omit"
                    f"\n{record}"
                )
                continue

            # Generate its URN
            c.urn = sdmx.urn.make(c)

    # Write to local store
    for codelist in cl.values():
        STORE.set(codelist)


def localize_all(
    value: str,
    translations: Mapping[str, "gettext.NullTranslations"],
    *,
    default_locale="en",
) -> dict[str, str]:
    """Localize `value` in all languages available in `translations`."""
    # Put the default locale first
    result = {default_locale: value}

    for lang, tr in translations.items():
        localized = tr.gettext(value)
        if localized != value:
            result[lang] = localized

    return result



"""Institute for Transport & Development Policy (ITDP) provider."""

from transport_data.util.pluggy import hookimpl


@hookimpl
def get_agencies():
    from sdmx.model import common

    a = common.Agency(
        id="ITDP",
        name="Institute for Transportation and Development Policy",
        contact=[common.Contact(email=["data@itdp.org"], uri=["https://itdp.org"])],
    )

    return (a,)



from typing import TYPE_CHECKING

import pytest
from sdmx.model import v21

from transport_data.ato import convert, dataset_to_metadata_reports, fetch
from transport_data.testing import MARK

if TYPE_CHECKING:
    import responses


@pytest.fixture(scope="session")
def ato_converted_data(mock_zenodo_api: "responses.RequestsMock", tmp_store) -> None:
    """Converted ATO data and structures in the test data directory."""
    # 'Proper' method: repeat the conversion in the test data directory
    for part in (
        # Temporarily disabled; see https://github.com/transport-data/tools/issues/48
        # "ACC",
        "APH",
        "CLC",
        "INF",
        "MIS",
        "RSA",
        "SEC",
        "TAS",
    ):
        with mock_zenodo_api:
            convert(part, from_zenodo=False)

    # # 'Fast' method: mirror the files from the user's directory
    # from shutil import copyfile

    # from transport_data import Config

    # source_dir = Config().data_path.joinpath("local")
    # dest_dir = tmp_store.store["local"].path

    # def predicate(p: Path) -> bool:
    #     return "ATO:" in p.name

    # for p in filter(predicate, source_dir.iterdir()):
    #     copyfile(p, dest_dir.joinpath(p.name))


@pytest.fixture
def ato_any_dataset(ato_converted_data) -> "v21.DataSet":  # pragma: no cover
    """One (any) ATO data set."""
    from transport_data import STORE

    key = STORE.list(v21.DataSet, maintainer="ATO")[0]
    return STORE.get(key)


@MARK[0]
def test_convert0(ato_converted_data):
    """Test that :func:`.ato.convert` works for certain parts."""
    # Nothing in particular: simply request the fixture that generates the parts
    # TODO Add assertions about the numbers of data structures and sets converted


@pytest.mark.parametrize(
    "part",
    (pytest.param("POL", marks=pytest.mark.xfail(reason="File format differs")),),
)
def test_convert1(mock_zenodo_api, part) -> None:
    """Test other `part`s that need particular marks."""
    with mock_zenodo_api:
        convert(part, from_zenodo=True)


@pytest.mark.parametrize("from_zenodo", (False, True))
def test_fetch(mock_zenodo_api, from_zenodo) -> None:
    with mock_zenodo_api:
        result = fetch(from_zenodo=from_zenodo, dry_run=True)

    # Expected number of file names available
    assert 9 == len(result)


@MARK[0]
@pytest.mark.usefixtures("ato_converted_data")
def test_dataset_to_metadata_reports():
    from transport_data import STORE
    from transport_data.org.metadata import get_msd

    # Retrieve one particular converted data set
    datasets = filter(
        lambda k: "TAS-VEP-001" in k, STORE.list(v21.DataSet, maintainer="ATO")
    )
    ds = STORE.get(next(datasets))

    # Conversion works
    result = dataset_to_metadata_reports(ds, get_msd())

    # Multiple metadata reports are generated
    assert 10 == len(result)



"""Utilities for :mod:`jinja2`."""

from functools import lru_cache


@lru_cache
def get_env():
    """Return a Jinja2 environment for rendering templates."""
    from jinja2 import Environment, PackageLoader, select_autoescape

    from transport_data import util
    from transport_data.org import metadata

    # Create a Jinja environment
    env = Environment(
        loader=PackageLoader("transport_data", package_path="data/template"),
        extensions=["jinja2.ext.loopcontrols"],
        autoescape=select_autoescape(),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    # Update filters
    def _get_reported_attribute(mdr, id_):
        for ra in mdr.metadata:
            if ra.value_for.id == id_:
                return ra.value, ra.value_for
        return "—", None

    def _format_desc(dim):
        try:
            anno_description = dim.get_annotation(id="tdc-description")
            if desc := str(anno_description.text):
                return desc
            else:
                raise KeyError
        except KeyError:
            return "—"

    env.filters["dfd_id"] = metadata.dfd_id
    env.filters["domain"] = util.url_hostname
    env.filters["format_desc"] = _format_desc
    env.filters["rst_heading"] = util.uline

    return env, dict(
        get_reported_attribute=_get_reported_attribute,
    )



"""CLI for :mod:`.proto`.

.. runblock:: console

   $ tdc proto --help

"""

import click


@click.command("proto")
def main():  # pragma: no cover
    """TDC prototype debugging."""
    # from transport_data import ato
    from transport_data import jrc

    # # commented: this fetches *all* data, and should not be done too often
    # # jrc.extract_all()
    #
    # # commented: uncomment these statements to fetch and unpack just one data file
    # path = jrc.fetch("AT")
    # jrc.extract(path)

    print(jrc.read("AT"))

    # # commented: uncomment this statement to fetch ATO ATO data files
    # ato.get()

    # ato.convert_all()



"""Utilities for :mod:`docutils`."""

from pathlib import Path
from typing import TYPE_CHECKING, cast
from zipfile import BadZipFile, ZipFile

import docutils.writers
import docutils.writers.odf_odt
from docutils.writers.odf_odt import SubElement

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element


class ODFTranslator(docutils.writers.odf_odt.ODFTranslator):
    """Translator from docutils DOM to OpenDocumentText

    This subclass works around bugs and missing figures in the upstream class.

    - Tolerate a file path with ".xml" for the "stylesheet" setting.
    - Fix internal hyperlinks:

      - Write internal hyperlink targets as :xml:`<text:bookmark>` instead of
        :xml:`<text:reference-mark>`.
      - Write hyperlinks to internal targets as :xml:`<text:a>` instead of
        :xml:`<text:reference-ref>`. The latter does not allow to specify the text
        content of the reference. In this class, the reference text is as given in the
        docutils DOM source.
    """

    # NB Lines excluded with "pragma: no cover" are not used in transport_data

    pending_ids: list[str]

    def retrieve_styles(self, extension: str) -> None:
        """Retrieve the stylesheet a file with extension either ".xml" or `extension`.

        Returns nothing.
        """
        from lxml import etree

        stylespath = Path(self.settings.stylesheet)

        if stylespath.suffix == ".xml":
            with open(stylespath, "rb") as stylesfile:
                self.str_stylesheet = stylesfile.read()
        elif stylespath.suffix == extension:  # pragma: no cover
            with ZipFile(stylespath) as zf:
                self.str_stylesheet = zf.read("styles.xml")
                self.str_stylesheetcontent = zf.read("content.xml")
        else:  # pragma: no cover
            raise RuntimeError(
                f"stylesheet path {stylespath} must be {extension} or .xml file"
            )

        self.dom_stylesheet = cast("Element", etree.fromstring(self.str_stylesheet))

        if self.str_stylesheetcontent:  # pragma: no cover
            # TODO Identify what this is for, or remove
            self.dom_stylesheetcontent = cast(
                "Element", etree.fromstring(self.str_stylesheetcontent)
            )
            self.table_styles = self.extract_table_styles(self.str_stylesheetcontent)

    def append_pending_ids(self, el) -> None:
        if self.settings.create_links:
            for id in self.pending_ids:
                SubElement(el, "text:bookmark", attrib={"text:name": id})
        self.pending_ids.clear()

    def visit_reference(self, node: "docutils.nodes.reference") -> None:
        # text = node.astext()
        if self.settings.create_links:
            if "refuri" in node:
                href = node["refuri"]
                if self.settings.cloak_email_addresses and href.startswith("mailto:"):
                    # href = self.cloak_mailto(href)
                    href = href
                el = self.append_child(
                    "text:a",
                    attrib={
                        "xlink:href": "%s" % href,
                        "xlink:type": "simple",
                    },
                )
                self.set_current_element(el)
            elif "refid" in node:
                el2 = self.append_child(
                    "text:a",
                    attrib={"xlink:type": "simple", "xlink:href": f"#{node['refid']}"},
                )
                el2.text = str(node.children.pop(0))
            else:  # pragma: no cover
                self.document.reporter.warning(
                    'References must have "refuri" or "refid" attribute.'
                )
        if (
            self.in_table_of_contents
            and len(node.children) >= 1
            and isinstance(node.children[0], docutils.nodes.generated)
        ):  # pragma: no cover
            node.remove(node.children[0])


class ODTWriter(docutils.writers.odf_odt.Writer):
    """Docutils writer for OpenDocument Text.

    This subclass works around bugs in the upstream class.
    """

    # NB Lines excluded with "pragma: no cover" are not used in transport_data

    def __init__(self):
        # These 2 statements are identical to the parent class' __init__, except using
        # the ODFTranslator subclass defined in this module
        docutils.writers.Writer.__init__(self)
        self.translator_class = ODFTranslator

        # Reconstruct default_stylesheet_path as an absolute (not relative) path
        self.default_stylesheet_path = Path(
            docutils.writers.odf_odt.__file__
        ).parent.joinpath(self.default_stylesheet)

    def copy_from_stylesheet(self, outzipfile: "ZipFile") -> None:
        """Copy images, settings, etc from the stylesheet doc into target doc."""
        stylespath = Path(self.settings.stylesheet)

        try:
            zf = ZipFile(stylespath)
        except BadZipFile:
            # `stylespath` references a non-ODF archive, which will not contain
            # settings.xml or any Pictures/ directory. Copy these instead from the
            # default file.
            zf = ZipFile(self.default_stylesheet_path)

        # Copy the styles
        self.write_zip_str(outzipfile, "settings.xml", zf.read("settings.xml"))

        # Copy the images
        for name in filter(
            lambda n: n.startswith("Pictures/"), zf.namelist()
        ):  # pragma: no cover
            outzipfile.writestr(name, zf.read(name))



from functools import partial
from pathlib import Path

import pytest

from transport_data.org.metadata import (
    contains_data_for,
    get_msd,
    groupby,
    merge_ato,
    report,
)
from transport_data.org.metadata.spreadsheet import make_workbook, read_workbook
from transport_data.testing import MARK
from transport_data.tests.test_ato import ato_converted_data  # noqa: F401

#: Number of metadata reports in the test specimen for which contains_data_for() returns
#: :any:`True`.
COUNTRIES = [
    ("CN", 19),
    ("ID", 17),
    ("IN", 19),
    ("PH", 14),
    ("NP", 0),
    ("TH", 17),
    ("VN", 18),
]


def test_get_msd():
    msd = get_msd()

    # MSD contains a report structure with id "ALL"
    rs = msd.report_structure["ALL"]

    # Report structure has a certain number of metadata attributes
    assert 9 == len(rs)

    # Metadata attributes have expected ID, annotation, and association to concepts with
    # expected properties
    mda = rs.get("DATA_DESCR")
    assert "<class 'str'>" == mda.eval_annotation(id="data-type")
    c = mda.concept_identity
    assert "DATA_DESCR" == c.id
    assert "Data description" == str(c.name)
    assert str(c.description).startswith("Any information about the data flow that ")


def test_make_workbook(tmp_path) -> None:
    make_workbook()


def test_make_workbook_cli(tmp_path, tdc_cli) -> None:
    with tdc_cli.isolated_filesystem(temp_dir=tmp_path) as td:
        # Expected output path
        exp = Path(td, "sample.xlsx")
        result = tdc_cli.invoke(["org", "template"])

    # Command ran without error
    assert 0 == result.exit_code

    # Expected file was generated
    assert exp.exists()


@MARK[0]
@pytest.mark.usefixtures("ato_converted_data")
def test_merge_ato(example_metadata) -> None:
    mds, cs = example_metadata

    merge_ato(mds)


@pytest.fixture(scope="function")
def example_metadata(test_data_path):
    """Example output from :func:`.read_workbook`.

    The fixture is function scoped, which allows that test code using the fixture
    alter the contents of the metadata set.
    """
    return read_workbook(test_data_path.joinpath("metadata-input.xlsx"))


def test_read_workbook(example_metadata) -> None:
    # Function runs successfully
    result, _ = example_metadata

    # Result has a certain number of metadata reports
    assert 45 == len(result.report)


@pytest.mark.parametrize("ref_area, N_exp", COUNTRIES)
def test_groupby(example_metadata, ref_area, N_exp: int) -> None:
    predicate = partial(contains_data_for, ref_area=ref_area)
    result = groupby(example_metadata[0], predicate)

    # Expected counts of metadata reports with respective values
    # NB Use set notation to tolerate missing keys in `result` if N_exp == 0
    exp = {(True, N_exp), (False, 45 - N_exp)}

    # Observed counts match
    assert exp >= {(k, len(v)) for k, v in result.items()}


class TestMetadataSet0ODT:
    def test_render(self, tmp_path, example_metadata) -> None:
        mds, cs_dims = example_metadata

        path = tmp_path.joinpath("all.odt")

        # Function runs successfully
        report.MetadataSet0ODT(mds).write_file(path)

        # Output was created
        assert path.exists()


class TestMetadataSet0Plain:
    def test_render(self, capsys, example_metadata) -> None:
        mds, cs_dims = example_metadata

        # Function runs successfully
        result = report.MetadataSet0Plain(mds).render()

        # pathlib.Path("debug.txt").write_text(result)  # DEBUG Write to a file
        # print(result)  # DEBUG Write to stdout

        # Output contains certain text
        assert "MEASURE: 39 unique values" in result

        # TODO expand with further assertions


class TestMetadataSet1HTML:
    @pytest.mark.parametrize("ref_area, N_exp", COUNTRIES)
    def test_write_file(self, tmp_path, example_metadata, ref_area, N_exp) -> None:
        path = tmp_path.joinpath(f"{ref_area}.html")

        report.MetadataSet1HTML(example_metadata[0], ref_area=ref_area).write_file(
            path, encoding="utf-8"
        )

        # Output was created
        assert path.exists()


@pytest.mark.parametrize("ref_area, N_exp", COUNTRIES)
class TestMetadataSet1ODT:
    def test_write_file(self, tmp_path, example_metadata, ref_area, N_exp) -> None:
        path = tmp_path.joinpath(f"{ref_area}.odt")

        report.MetadataSet1ODT(example_metadata[0], ref_area=ref_area).write_file(
            path=path
        )

        # Output was created
        assert path.exists()


class TestMetadataSet2HTML:
    def test_write_file(self, tmp_path, example_metadata) -> None:
        path = tmp_path.joinpath("all.html")

        report.MetadataSet2HTML(
            example_metadata[0], ref_area=list(item[0] for item in COUNTRIES)
        ).write_file(path, encoding="utf-8")

        # Output was created
        assert path.exists()


def test_read_cli(tmp_path, tdc_cli, test_data_path) -> None:
    path_in = test_data_path.joinpath("metadata-input.xlsx")

    result = tdc_cli.invoke(["org", "read", str(path_in)])

    # Command ran without error
    assert 0 == result.exit_code, result.output

    assert "MEASURE: 39 unique values" in result.output


def test_refresh_cli(tdc_cli) -> None:
    result = tdc_cli.invoke(["org", "refresh"])

    # Command ran without error
    assert 0 == result.exit_code, result.output


def test_summarize_cli(tmp_path, tdc_cli, test_data_path) -> None:
    path_in = test_data_path.joinpath("metadata-input.xlsx")

    # --ref-area= single value
    with tdc_cli.isolated_filesystem(temp_dir=tmp_path) as td:
        path_out = Path(td, "VN")
        result = tdc_cli.invoke(["org", "summarize", "--ref-area=VN", str(path_in)])

    # Command ran without error
    assert 0 == result.exit_code, result.output
    # Output files are generated
    assert path_out.with_suffix(".html").exists()
    assert path_out.with_suffix(".odt").exists()

    # --ref-area= multiple values
    with tdc_cli.isolated_filesystem(temp_dir=tmp_path) as td:
        path_out = Path(td, "all")
        result = tdc_cli.invoke(["org", "summarize", "--ref-area=TH,VN", str(path_in)])

    # Command ran without error
    assert 0 == result.exit_code, result.output
    # Output files are generated
    assert path_out.with_suffix(".html").exists()
    assert path_out.with_suffix(".odt").exists()


def test_tuewas_cli(tmp_path, tdc_cli, test_data_path) -> None:
    path_in = test_data_path.joinpath("metadata-input.xlsx")
    with tdc_cli.isolated_filesystem(temp_dir=tmp_path) as td:
        dir_out = Path(td, "output")
        result = tdc_cli.invoke(["org", "tuewas", str(path_in)])

    # Command ran without error
    assert 0 == result.exit_code, result.output
    # Expected files were generated
    assert dir_out.joinpath("Metadata summary.odt").exists()
    assert dir_out.joinpath("Metadata summary table.html").exists()
    assert dir_out.joinpath("CN", "Summary.odt").exists()



from functools import partial
from shutil import copyfile
from typing import TYPE_CHECKING

import pytest
import sdmx
from sdmx.message import DataMessage
from sdmx.model import v21

from transport_data import Config, testing
from transport_data.itdp.rtdb import convert, fetch

if TYPE_CHECKING:
    import pathlib


@pytest.fixture(scope="session")
def rtdb_test_data(test_data_path, tmp_config: Config) -> "pathlib.Path":
    """Copy of the RTDB snapshot in the test cache directory."""
    source = test_data_path.joinpath("itdp", "rtdb.xlsx")
    dest = tmp_config.cache_path.joinpath("itdp", "rtdb.xlsx")
    dest.parent.mkdir(parents=True, exist_ok=True)

    copyfile(source, dest)

    return dest


COUNTRIES = [
    ("CN", 144),
    ("ID", 43),
    ("IN", 97),
    ("PH", 78),
    ("NP", 0),
    ("TH", 58),
    ("VN", 13),
]


def test_convert(caplog, rtdb_test_data: "pathlib.Path") -> None:
    ds1 = convert(rtdb_test_data)

    def _filter_geo(value: str, obs: "v21.Observation") -> bool:
        assert obs.dimension
        return obs.dimension["GEO"].value == value

    for geo, N in COUNTRIES:
        # Create a DataMessage to work around https://github.com/khaeru/sdmx/issues/251
        dfd = v21.DataflowDefinition(structure=ds1.structured_by)  # type: ignore [arg-type]

        ds2 = v21.DataSet(
            described_by=dfd,
            structured_by=ds1.structured_by,
            obs=list(filter(partial(_filter_geo, geo), ds1.obs)),
        )
        dm = DataMessage(dataflow=dfd, data=[ds2])

        if not len(ds2):
            continue

        df = sdmx.to_pandas(dm).reset_index()

        assert N == len(df.query("value > 0 "))


@pytest.mark.xfail(
    condition=testing.GITHUB_ACTIONS,
    reason="No Google Cloud API credentials on GitHub Actions",
    raises=RuntimeError,
)
def test_fetch():
    fetch()



from pathlib import Path

import pandas as pd
import sdmx

from transport_data.iamc import structures_for_data, variable_cl_for_dsd


def test_structures_for_data():
    # Load test data
    test_data_path = Path(__file__).parents[1].joinpath("data", "tests")
    df = pd.read_csv(test_data_path.joinpath("iamc.csv"))

    # Function runs, returns a SDMX StructureMessage containing multiple structure
    # objects
    sm = structures_for_data(df, base_id="TEST")

    # Code lists have expected length
    expected = {
        "MODEL": 1,
        "SCENARIO": 6,
        "REGION": 10,
        "UNIT": 39,
    }

    for name, N in expected.items():
        assert N == len(sm.codelist[name])

    # Concept schemes have expected length
    assert 31 == len(sm.concept_scheme["MEASURE"])

    # DSD is annotated with a description of the data from which it was derived
    dsd = sm.structure["TEST"]
    assert str(dsd.description).startswith("The original data are in")


def test_variables_cl_for_dsd(tmp_path, sdmx_structures):
    # Function runs on the "MASS" DSD
    cl = variable_cl_for_dsd(sdmx_structures.structure["MASS"])

    # Expected number of variable names is generated
    assert 25 == len(cl)

    # Function runs on the "PICKED" DSD
    variable_cl_for_dsd(sdmx_structures.structure["PICKED"], cl)

    # Items are appended to `cl`
    assert 50 == len(cl)

    # Codelist can be roundtripped via file
    path = tmp_path.joinpath("data.xml")
    # print(f"{path = }")

    sm_out = sdmx.message.StructureMessage()
    sm_out.add(cl)

    with open(path, "wb") as f:
        f.write(sdmx.to_xml(sm_out, pretty_print=True))

    sm_in = sdmx.read_sdmx(path)
    cl_in = sm_in.codelist["VARIABLE"]

    # Codes can be accessed by ID == IAMC variable name
    code = cl_in["Mass of fruit|Green"]

    # Code is annotated with the URN of the full-dimensionality DSD
    assert code.eval_annotation("iamc-full-dsd").endswith(
        "DataStructureDefinition=TEST:MASS(1.0.0)"
    )
    # Code is annotated with its full key; no missing dimensions
    assert {"COLOUR": "GREEN", "FRUIT": "_T"} == code.eval_annotation("iamc-full-key")



from typing import TYPE_CHECKING

from platformdirs import user_data_path

if TYPE_CHECKING:
    import googleapiclient.discovery


def get_service(
    *args, scopes: list[str], **kwargs
) -> "googleapiclient.discovery.Resource":
    from google.auth.transport.requests import Request
    from google.oauth2.credentials import Credentials
    from google_auth_oauthlib.flow import InstalledAppFlow
    from googleapiclient.discovery import build

    udp = user_data_path("transport-data")
    credentials_path = udp.joinpath("google-cloud-credentials.json")
    token_path = udp.joinpath("google-cloud-token.json")

    creds = None

    # The file token.json stores the user's access and refresh tokens, and is created
    # automatically when the authorization flow completes for the first time.
    if token_path.exists():
        creds = Credentials.from_authorized_user_file(token_path, scopes)

    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            try:
                flow = InstalledAppFlow.from_client_secrets_file(
                    credentials_path, scopes
                )
            except FileNotFoundError:
                raise RuntimeError(f"No Google API credentials at {credentials_path}")
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        token_path.write_text(creds.to_json())

    kwargs.update(credentials=creds)
    return build(*args, **kwargs)



"""Generate reports about TDC-structured metadata.

:class:`.Report` subclasses in this file **should** have names like::

    { Type }{ ID }{ Format }

…wherein:

- ``{ Type }`` refers to the type of object(s) from the SDMX Information Model that
  is/are represented in the report.
  Usually the first argument to the :py:`__init__()` method is an instance of this type.
- ``{ ID }`` is a number that distinguishes different ‘kinds’ of reports for the same
  ``{ Type }``.
  Report classes with the same ``{ Type }{ ID }`` **should** display roughly the same
  information in the same order and layout, regardless of ``{ Format }``.
- ``{ Format }`` is the output format or file type.
"""

from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING

from transport_data.report import Report
from transport_data.util import libreoffice, uline

if TYPE_CHECKING:
    import pathlib

    from sdmx.model import v21


@dataclass
class MetadataAttribute0Plain(Report):
    """Unique values appearing in `mds` for metadata attribute `mda_id`.

    Each unique value is shown with the IDs of the data flows that contain the value for
    `mda_id`.
    """

    #: Metadata set to report.
    mds: "v21.MetadataSet"
    #: ID of a Metadata Attribute to report.
    mda_id: str

    def render(self) -> str:
        from transport_data.org.metadata import map_values_to_ids

        value_id = map_values_to_ids(self.mds, self.mda_id)

        assert self.mds.structured_by
        mda = self.mds.structured_by.report_structure["ALL"].get(self.mda_id)

        lines = ["", "", uline(f"{mda}: {len(value_id)} unique values")]

        for value, df_ids in sorted(value_id.items()):
            lines.extend([value, "    " + " ".join(sorted(df_ids))])

        return "\n".join(lines)


@dataclass
class MetadataAttribute0RST(Report):
    """Unique values appearing in `mds` for the metadata attribute `mda_id`.

    Same as :class:`.MetadataAttribute0Plain`, but in reStructuredText.
    """

    #: Jinja2 reStructuredText template.
    template_name = "metadata-attribute-0.rst"
    #: Metadata set to report.
    mds: "v21.MetadataSet"
    #: ID of a Metadata Attribute to report.
    mda_id: str

    def render(self) -> str:
        from transport_data.org.metadata import map_values_to_ids

        value_id = map_values_to_ids(self.mds, self.mda_id)
        assert self.mds.structured_by
        mda = self.mds.structured_by.report_structure["ALL"].get(self.mda_id)

        return self.render_jinja_template(mda=mda, value_id=value_id)


@dataclass
class MetadataReport0HTML(Report):
    """Same as :class:`.MetadataReport0Plain`, but in HTML."""

    #: Metadata report to report.
    mdr: "v21.MetadataReport"

    def render(self) -> str:
        raise NotImplementedError


@dataclass
class MetadataReport0Plain(Report):
    """Contents of a single metadata report.

    This includes:

    1. The reported attribute values for all metadata attributes.
    2. The data flow that is targeted by the report and its dimensions.
    """

    #: Metadata report to report.
    mdr: "v21.MetadataReport"

    def render(self) -> str:
        lines = ["", uline("Metadata report")]

        # Retrieve references to the data flow and data structure
        dfd: "v21.DataflowDefinition" = self.mdr.attaches_to.key_values["DATAFLOW"].obj  # type: ignore [union-attr]
        dsd = dfd.structure

        # Summarize the data flow and data structure

        lines.extend(
            [f"Refers to {dfd!r}", f"  with structure {dsd!r}", "    with dimensions:"]
        )
        for dim in dsd.dimensions:
            line = f"    - {dim.id}:"
            try:
                anno_description = dim.get_annotation(id="tdc-description")
                if desc := str(anno_description.text):
                    line += f" {desc!s}"
                else:
                    raise KeyError
            except KeyError:
                line += " (no info)"

            try:
                original_id = dim.get_annotation(id="tdc-original-id").text
                line += f" ('{original_id!s}' in input file)"
            except KeyError:
                pass
            lines.append(line)

        lines.append("")

        for ra in self.mdr.metadata:
            if ra.value_for.id == "DATAFLOW":
                continue
            assert hasattr(ra, "value")
            lines.append(f"{ra.value_for}: {ra.value}")

        return "\n".join(lines)


@dataclass
class MetadataSet0ODT(Report):
    """Summary of the unique reported attribute values in `mds`.

    Similar to :class:`.MetadataSet0Plain`, but also including:

    - The unique dimension concepts appearing in the data structure definitions.
    """

    template_name = "metadata-set-0.rst"
    #: Metadata set to report.
    mds: "v21.MetadataSet"

    def render(self) -> bytes:
        from transport_data.org.metadata import map_dims_to_ids

        # Mapping from reported attribute values to data flow IDs
        mda = [
            MetadataAttribute0RST(self.mds, "DATA_PROVIDER").render(),
            MetadataAttribute0RST(self.mds, "MEASURE").render(),
            MetadataAttribute0RST(self.mds, "UNIT_MEASURE").render(),
        ]
        # Mapping from dimension IDs to data flow IDs
        dim_id = map_dims_to_ids(self.mds)

        rst_source = self.render_jinja_template(mda=mda, dim_id=dim_id)
        # print(rst_source)  # DEBUG
        return self.rst2odt(rst_source)

    def write_file(self, path: "pathlib.Path", **kwargs) -> None:
        """:meth:`render` the report and write to `path`."""
        super().write_file(path, **kwargs)
        libreoffice.to_pdf(path)


@dataclass
class MetadataSet0Plain(Report):
    """Summary of the unique reported attribute values in `mds`.

    This includes:

    - Unique values of the metadata attributes ``MEASURE``, ``DATA_PROVIDER``, and
      ``UNIT_MEASURE``.
    """

    #: Metadata set to report.
    mds: "v21.MetadataSet"

    def render(self) -> str:
        lines = [
            f"Metadata set containing {len(self.mds.report)} metadata reports",
            MetadataAttribute0Plain(self.mds, "MEASURE").render(),
            MetadataAttribute0Plain(self.mds, "DATA_PROVIDER").render(),
            MetadataAttribute0Plain(self.mds, "UNIT_MEASURE").render(),
        ]

        for r in self.mds.report:
            lines.append(MetadataReport0Plain(r).render())

        return "\n".join(lines)


@dataclass
class MetadataSet1HTML(Report):
    """Metadata reports related to `ref_area`."""

    template_name = "metadata-set-1.html"

    #: Metadata set to summarize.
    mds: "v21.MetadataSet"
    #: Geography.
    ref_area: str

    def render(self) -> str:
        from transport_data.org.metadata import contains_data_for, groupby

        grouped = groupby(
            self.mds, key=partial(contains_data_for, ref_area=self.ref_area)
        )

        return self.render_jinja_template(
            ref_area=self.ref_area, matched=grouped[True], no_match=grouped[False]
        )


@dataclass
class MetadataSet1ODT(Report):
    """Metadata reports related to `ref_area`.

    Same as :class:`.MetadataSet1HTML` but as OpenDocument Text.
    """

    template_name = "metadata-set-1.rst"

    #: Metadata set to summarize.
    mds: "v21.MetadataSet"
    #: Geography.
    ref_area: str

    def render(self) -> bytes:
        from transport_data.org.metadata import contains_data_for, groupby

        # Prepare data; same as SummaryHTML0
        grouped = groupby(
            self.mds, key=partial(contains_data_for, ref_area=self.ref_area)
        )

        # Render the report as reStructuredText using Jinja2 and a template
        rst_source = self.render_jinja_template(
            ref_area=self.ref_area,
            matched=grouped[True],
            no_match=grouped[False],
            show_no_match=False,
        )
        # with open("debug.rst", "w") as f: # DEBUG
        #     f.write(rst_source)
        # print(rst_source)  # DEBUG

        # Convert reStructuredText → OpenDocumentText
        return self.rst2odt(rst_source)

    def write_file(self, path: "pathlib.Path", **kwargs) -> None:
        """:meth:`render` the report and write to `path`."""
        super().write_file(path, **kwargs)
        libreoffice.to_pdf(path)


@dataclass
class MetadataSet2HTML(Report):
    """Table of metadata reports.

    This table has:

    - One *column* per value in `ref_areas`.
    - One *row* per metadata report in `mds`.
    - A check-mark in cells where :func:`.contains_data_for` indicates that the metadata
      report targets a data flow that contains data for the reference area.
    """

    template_name = "metadata-set-2.html"

    #: Metadata set to summarize.
    mds: "v21.MetadataSet"
    #: Geographies to show.
    ref_area: list[str]

    def render(self) -> str:
        from transport_data.org.metadata import contains_data_for

        data = {
            mdr.attaches_to.key_values["DATAFLOW"].obj.id: {  # type: ignore [union-attr]
                ra: contains_data_for(mdr, ra) for ra in self.ref_area
            }
            for mdr in self.mds.report
        }

        return self.render_jinja_template(ref_area=self.ref_area, data=data)



import logging
from typing import TYPE_CHECKING

import pytest
from click.testing import CliRunner
from sdmx.model import common, v21

from transport_data.config import Config
from transport_data.org import get_agencyscheme
from transport_data.store import main

if TYPE_CHECKING:
    import dsss.store

log = logging.getLogger(__name__)


class TestBaseStore:
    """:mod:`dsss` classes provide features required by :mod:`transport_data`."""

    @pytest.fixture
    def s(self, tmp_config: Config, sdmx_structures) -> "dsss.store.Store":
        from dsss.store import DictStore

        # Create a DictStore
        s = DictStore()

        # Set some items
        s.set(sdmx_structures.codelist["FRUIT"])

        return s

    def test_list_versions(self, s: "dsss.store.Store") -> None:
        result = s.list_versions(common.Codelist, maintainer="TEST", id="FRUIT")

        assert ("1.0.0",) == result

        assert tuple() == s.list_versions(common.Codelist, maintainer="TEST", id="FOO")

    def test_assign_version(self, sdmx_structures, s: "dsss.store.Store") -> None:
        cl: "common.Codelist" = common.Codelist(
            id="FRUIT", maintainer=common.Agency(id="TEST")
        )

        s.assign_version(cl, major=1)

        assert "2.0.0" == cl.version


class TestUnionStore:
    """Test the :class:`transport_data.store.UnionStore` subclass."""

    @pytest.fixture(scope="class")
    def tdci_agencyscheme(self, tmp_store):
        tmp_store.set(get_agencyscheme())

    @pytest.mark.usefixtures("sdmx_structures")
    def test_add_to_registry(self, tmp_store) -> None:
        tmp_store.add_to_registry("Codelist=TEST:FRUIT(1.0.0)")

    @pytest.mark.usefixtures("sdmx_structures", "tdci_agencyscheme")
    @pytest.mark.parametrize(
        "urn, cls",
        (
            (
                "urn:sdmx:org.sdmx.infomodel.base.AgencyScheme=TDCI:TDCI",
                common.AgencyScheme,
            ),
            (
                "urn:sdmx:org.sdmx.infomodel.base.AgencyScheme=TDCI:TDCI(2025.12.27)",
                common.AgencyScheme,
            ),
            ("AgencyScheme=TDCI:TDCI(2025.12.27)", common.AgencyScheme),
            ("AgencyScheme=TDCI:TDCI", common.AgencyScheme),
            # DSD can be retrieved with normalized or non-normalized ID
            ("DataStructure=TEST:MASS(1.0.0)", v21.DataStructureDefinition),
            ("DataStructureDefinition=TEST:MASS(1.0.0)", v21.DataStructureDefinition),
        ),
    )
    def test_get(self, tmp_store, urn, cls) -> None:
        """Objects can be retrieved by partial URN."""
        try:
            result = tmp_store.get(urn)
        except Exception:  # pragma: no cover
            log.error("\n".join(["Not found among:"] + tmp_store.list(cls)))
            raise
        assert isinstance(result, cls)

    def test_list(self, tmp_store) -> None:
        result = sorted(tmp_store.list(maintainer="TEST"))
        assert 5 <= len(result)
        # Some known entries are present
        assert {
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=TEST:COLOUR(1.0.0)",
            "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=TEST:PICKED(1.0.0)",
        } <= set(result)


def test_cli_list(sdmx_structures):
    result = CliRunner().invoke(main, ["list", "--maintainer=TEST"])
    assert 0 == result.exit_code
    # FIXME Avoid a result that depends on test run order
    assert 4 <= len(result.output.splitlines()) <= 8, result.output


def test_cli_show(sdmx_structures):
    result = CliRunner().invoke(main, ["show", "Codelist=TEST:COLOUR(1.0.0)"])
    assert 0 == result.exit_code
    assert "  4 <Code _T: Total>" in result.output



""":class:`.ckan.Client` objects to access TDC instances."""

import typing
from dataclasses import dataclass
from functools import lru_cache
from typing import TYPE_CHECKING, Union

import click

from transport_data.util.ckan import Client, ModelProxy, Package
from transport_data.util.sdmx import fields_to_mda

if TYPE_CHECKING:
    import uuid

    from sdmx.model import v21

#: Development instance of TDC CKAN. Accessible as of 2024-11-27.
DEV = Client("https://ckan.tdc.dev.datopian.com", id="dev")

#: Production instance of TDC CKAN. Accessible as of 2024-11-27.
PROD = Client("https://ckan.transport-data.org", id="prod")

#: Staging instance of TDC CKAN. Not accessible as of 2024-11-27; SSL certificate error.
STAGING = Client("https://ckan.tdc.staging.datopian.com", id="staging")


@dataclass
class CKANMetadataReportStructure:
    """Concepts in the CKAN metadata for a :class:`Package`."""

    #: JSON data
    #:
    #: This is optional. If used, it contains the JSON content returned by the CKAN API.
    #: This **must** be identical to the values appearing in the other fields.
    JSON: dict

    author: str | None
    author_email: str | None
    contributors: list
    creator_user_id: "uuid.UUID"
    data_access: str
    data_provider: str
    dimensioning: str
    frequency: str
    geographies: list
    groups: list
    id: "uuid.UUID"
    indicators: list
    is_archived: bool
    isopen: bool
    language: str
    license_id: str
    license_title: str
    license_url: str
    maintainer: str | None
    maintainer_email: str | None
    metadata_created: str
    metadata_modified: str
    modes: list
    name: str
    notes: str
    num_resources: int
    num_tags: int
    organization: list
    owner_org: "uuid.UUID"
    private: bool
    regions: list
    relationships_as_object: list
    relationships_as_subject: list
    resources: list
    sectors: list
    services: list
    sources: list
    state: str
    tags: list
    tdc_category: str
    temporal_coverage_end: str
    temporal_coverage_start: str
    title: str
    type: str
    units: list
    url: str
    version: str | None


#: :class:`click.Option` for :program:`--instance/-i` that allows the user to select one
#: of :data:`PROD` (default), :data:`DEV`, or :data:`STAGING`.
#:
#: A function using this option gets a `instance` keyword argument that is the
#: appropriate :class:`.Client` instance. :attr:`.Client.id` can be inspected for the
#: given or default option value.
instance_option = click.Option(
    ["--instance", "-i"],
    type=click.Choice(["dev", "prod", "staging"]),
    default="prod",
    callback=lambda c, p, v: {"dev": DEV, "prod": PROD, "staging": STAGING}[v],
    help="TDC CKAN instance to query.",
)


@lru_cache
def get_msd() -> "v21.MetadataStructureDefinition":
    """Generate and return an MSD containing all of the CKAN fields."""
    from sdmx.model import common, v21

    from transport_data import STORE
    from transport_data.org import get_agencyscheme

    TDCI = get_agencyscheme()["TDCI"]

    cs = common.ConceptScheme(id="CKAN_METADATA_CONCEPTS", maintainer=TDCI)
    msd = v21.MetadataStructureDefinition(id="CKAN", version="1", maintainer=TDCI)
    rs = msd.report_structure["ALL"] = v21.ReportStructure(id="ALL")

    fields_to_mda(CKANMetadataReportStructure, rs, cs)

    STORE.set(msd)
    STORE.set(cs)

    return msd


def ckan_package_to_mdr(p) -> "v21.MetadataReport":
    """Convert a :class:`.Package` instance to a MetadataReport."""
    from sdmx.model import v21

    msd = get_msd()

    mdr = v21.MetadataReport()

    ONEAV = v21.OtherNonEnumeratedAttributeValue

    for mda in msd.report_structure["ALL"]:
        av = ONEAV(value_for=mda)
        if mda.id == "JSON":
            av.value = repr(p.asdict())
        else:
            value = getattr(p, mda.id)
            av.value = value if isinstance(value, str) else repr(value)
        mdr.metadata.append(av)

    return mdr


def mdr_to_ckan_package(mdr):
    """Convert a MetadataReport to a :class:`.Package` instance."""
    data = {}
    # Iterate over reported attributes
    for ra in mdr.metadata:
        mda = ra.value_for  # The MetadataAttribute for which this

        # Don't store the raw JSON
        # TODO Branch within this function: either use the "JSON" attribute directly and
        #      ignore all other reported attributes, or vice versa
        if mda.id == "JSON":
            continue

        # Determine the data type from a "data-type" annotation on the MetadataAttribute
        dt_anno = mda.eval_annotation(id="data-type", globals=dict(typing=typing))
        data_type = {
            "<class 'bool'>": bool,
            "<class 'int'>": int,
            "<class 'list'>": list,
            "<class 'str'>": str,
            "uuid.UUID": str,
        }.get(dt_anno, dt_anno)

        # Convert the string representation of the value to its Python/JSON data
        # structure
        if data_type is None:
            raise NotImplementedError
        elif data_type in (bool, list):
            data[ra.value_for.id] = eval(ra.value)
        elif data_type == Union[str, None]:
            data[ra.value_for.id] = None if ra.value == "None" else str(ra.value)
        else:
            data[ra.value_for.id] = data_type(ra.value)

    return Package(data)


@click.command("ckan", params=[instance_option])
@click.option("--verbose", "-v", is_flag=True, help="Show more output.")
@click.argument("action")
@click.argument("args", nargs=-1)
def main(instance: Client, verbose, action, args):
    """Interact with the TDC CKAN instance via its API.

    ACTION is the ID of one of endpoints of the CKAN Action API, for instance "tag_show"
    or "package_list". ARGS are a space-separated list of key=value pairs, for instance
    "sort=title limit=10".

    See https://docs.ckan.org/en/latest/api/ for the list and further documentation, or
    use the "help_show" endpoint, for example:

        tdc ckan help_show name=package_search

    For some endpoints, the output is formatted. For all others, the raw JSON response
    is shown.
    """
    client = instance

    # Parse key=value args to dictionary
    data_dict = {key: value for key, _, value in map(lambda s: s.partition("="), args)}

    try:
        result = getattr(client, action)(**data_dict)
    except Exception as e:
        raise click.ClickException(repr(e))

    if action.endswith("_list"):
        for i, obj in enumerate(result, start=1):
            print(f"{i:3d}. {obj}")
    elif isinstance(result, str):
        print(result)
    else:  # pragma: no cover
        print(repr(result))
        if verbose and isinstance(result, ModelProxy):
            print(result.asdict())



"""International Organization of Motor Vehicle Manufacturers (OICA) provider.

This module handles data from the `OICA website <https://www.oica.net>`__. "Handle"
includes:

- Fetch data files.
- Parse the native spreadsheet layout. This layout differs across data flows and years;
  currently only the most recent formats are supported.
- Convert to SDMX following TDCI conventions.
"""

import logging
import re
from collections.abc import Iterator
from functools import lru_cache, partial
from itertools import count, product
from operator import itemgetter
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd

from transport_data.util.pluggy import hookimpl
from transport_data.util.pooch import Pooch

if TYPE_CHECKING:
    import sdmx.model.common

log = logging.getLogger(__name__)

#: Base URL for data files.
BASE_URL = "https://www.oica.net/wp-content/uploads/2025/"

REGISTRY_FILE = Path(__file__).with_name("registry.txt")

#: Pooch instance with registry populated from the file :data:`REGISTRY_FILE`.
POOCH = Pooch(module=__name__, base_url=BASE_URL, registry=REGISTRY_FILE)


def convert(
    measure: str,
) -> dict[str, "sdmx.model.v21.DataSet"]:
    """Convert OICA spreadsheets to SDMX.

    This method handles OICA's particular arrangement of data from a single data flow
    across multiple Microsoft Excel files, each with a single sheet. The individual
    files/sheets are processed by :func:`.convert_single_file`.

    The OICA files each mix data for multiple data flows or measures: for instance, each
    of the "Vehicles in use" files contains data for up to 3 different measures:

    - ``STOCK``: Number of vehicles in use.
    - ``STOCK_AAGR``: Average annual growth rate across a specified ``TIME_PERIOD``.
    - (in some cases) ``STOCK_CAP``: Number of vehicles in use per 1000 capita
      population.

    This function returns a dictionary in which data flow IDs for each measure are keys,
    and values are SDMX data sets, each for a single measure.
    """
    # Retrieve the DFD
    dfd, dsd = get_structures(measure)

    if "PROD" in dfd.id:
        raise NotImplementedError(f"Convert OICA data files for {dfd}")

    # Convert all fetchable/fetched files related to the DFD
    results = [convert_single_file(path, dfd) for path in filenames_for_dfd(dfd)]

    # Merge observations from multiple files into a single data set per data flow
    result: dict[str, "sdmx.model.v21.DataSet"] = {}
    for r in results:
        for dfd_id, datasets in r.items():
            if dfd_id not in result:
                result[dfd_id] = datasets.pop(0)
            for ds in datasets:
                result[dfd_id].add_obs(ds.obs)

    return result


def convert_tp(time_period: str, units: str, vehicle_type: str):
    """Identify values for 4 concepts using column labels melted into TIME_PERIOD."""
    concepts = ["MEASURE", "UNIT_MEASURE", "VEHICLE_TYPE", "TIME_PERIOD"]

    def _(*values):
        return pd.Series(dict(zip(concepts, values)))

    if time_period in (2015.0, 2020.0):
        return _("STOCK", units, vehicle_type, str(int(time_period)))
    elif time_period == "Average Annual Growth Rate\n 2015-2020":
        return _("STOCK_AAGR", "1", vehicle_type, "2015–2020")
    elif time_period == "Number of vehicles per 1000 inhabitants":
        return _("STOCK_CAP", "0.001", vehicle_type, "_X")
    elif match := re.fullmatch(r"Q1-Q4 (\d{4})", time_period):
        return _("SALES", units, vehicle_type, match.group(1))
    elif match := re.fullmatch(r"(2\d{3})/\s*(2\d{3})", time_period):
        return _("SALES_GR", "1", vehicle_type, f"{match.group(2)}–{match.group(1)}")
    elif time_period == "Scope":
        # log.warning("Discard values for 'Scope' attribute")
        return _(None, None, None, None)
    else:  # pragma: no cover
        raise ValueError(time_period)


def convert_single_file(
    path: Path, dfd: "sdmx.model.v21.DataflowDefinition"
) -> dict[str, list["sdmx.model.v21.DataSet"]]:
    """Convert single OICA data spreadsheet to SDMX.

    This function currently handles the 2020 and later file format used for sales and
    stock data.
    """
    from sdmx.model.v21 import DataSet

    from transport_data import STORE
    from transport_data.util.sdmx import make_obs

    if path.stem.endswith("2018") or path.stem.endswith("2019"):
        log.warning("Skip not implemented 2018–2019 OICA file format")
        return dict()

    # - Read data
    # - Drop entirely empty rows.
    # - Drop entirely empty columns.
    df = (
        pd.read_excel(path)
        .dropna(axis=0, how="all", ignore_index=True)  # type: ignore [call-overload]
        .dropna(axis=1, how="all")
    )

    # Confirm vehicle type and measure vs. contents of the title cell
    if "SALES" in dfd.id:
        pattern = r"REGISTRATIONS OR SALES OF NEW VEHICLES - (ALL TYPES|COMMERCIAL VEHICLES|PASSENGER CARS)"
        units = "vehicle"
    elif "STOCK" in dfd.id:
        pattern = r"(PC|CV|TOTAL) WORLD VEHICLES IN USE +\(in thousand units\)"
        units = "kvehicle"

    if match := re.fullmatch(pattern, str(df.iloc[0, 0])):
        # Codes are inconsistent across files; use shorter codes
        vehicle_type = {
            "ALL TYPES": "_T",
            "COMMERCIAL VEHICLES": "CV",
            "PASSENGER CARS": "PC",
            "TOTAL": "_T",
        }.get(match.group(1), match.group(1))
    else:  # pragma: no cover
        raise ValueError(f"Unrecognized table header: {df.iloc[0, 0]!r}")

    # - Drop title cell.
    # - Set index.
    # - Rename REGIONS/COUNTRIES to GEO.
    # - Melt to long format.
    # - Replace float with string year values.
    df = (
        df.iloc[2:, :]
        .set_axis(df.iloc[1, :], axis=1)
        .rename(columns={"REGIONS/COUNTRIES": "GEO"})
        .melt(id_vars="GEO", var_name="TIME_PERIOD", value_name="OBS_VALUE")
        .dropna(subset=["GEO", "OBS_VALUE"], how="any")
    )

    # Prepare a GEO codelist and map using the "GEO" column
    cl_geo = get_cl_geo()
    geo_map = _make_geo_codes(
        cl_geo, df["GEO"], maintainer=get_agencies()[0], version="0.1"
    )
    # Store `cl_geo`, including any added entries
    STORE.set(cl_geo)

    # Transform data
    # - Replace GEO values with codes from `cl_geo`.
    # - Replace TIME_PERIOD values with new TIME_PERIOD, MEASURE, UNIT_MEASURE,
    #   VEHICLE_TYPE.
    # - Preserve OBS_VALUE.
    df = pd.concat(
        [
            df["GEO"].replace(geo_map),
            df["TIME_PERIOD"].apply(convert_tp, units=units, vehicle_type=vehicle_type),
            df["OBS_VALUE"],
        ],
        axis=1,
    ).dropna(subset=["MEASURE", "VEHICLE_TYPE"], how="any")

    result: dict[str, list[DataSet]] = {}

    # Group data by MEASURE ~ dataflow
    for m, group_df in df.groupby("MEASURE"):
        # Create structures for this measure
        # TODO Retrieve possibly-cached or -stored structures
        dfd, dsd = get_structures(str(m))

        # Create a data set
        ds = DataSet(described_by=dfd, structured_by=dsd)

        # Convert rows of `group_df` to observations
        ds.add_obs(
            map(partial(make_obs, dsd=dsd), map(itemgetter(1), group_df.iterrows()))
        )

        # Store the data set
        result.setdefault(dfd.id, [])
        result[dfd.id].append(ds)

    return result


def fetch(dry_run: bool = False):
    """Fetch all known OICA data files."""
    if dry_run:
        for f in POOCH.registry:
            print(f"Valid url for GEO={f}: {POOCH.is_available(f)}")
        return

    return [POOCH.fetch(f) for f in POOCH.registry]


def filenames_for_dfd(
    dfd: "sdmx.model.v21.DataflowDefinition", fetch: bool = True
) -> Iterator[Path]:
    """Yield OICA's file names for individual files in `dfd`.

    Parameters
    ----------
    fetch : bool, optional
        If :obj:`.True` (the default), yield only files that are fetched and available
        locally. Otherwise, yield all possible names conforming to a known pattern.
    """
    if "PROD" in dfd.id:
        pattern = "{}-{}.xlsx"
        values = (
            {
                "By-country-region",
                "Passenger-Cars",
                "Light-Commercial-Vehicles",
                "Heavy-Trucks",
                "Buses-and-Coaches",
            },
            range(2017, 2024),
        )
    elif "SALES" in dfd.id:
        pattern = "{}_sales_{}.xlsx"
        values = {"cv", "pc", "total"}, range(2017, 2025)
    elif "STOCK" in dfd.id:
        pattern = "{}-World-vehicles-in-use-{}.xlsx"
        values = {"CV", "PC", "Total"}, range(2017, 2024)
    else:  # pragma: no cover
        raise NotImplementedError(f"Construct OICA file names for {dfd}")

    # Generate a cartesian product of the iterables in `values`
    for name_parts in product(*values):
        # Format a file name
        name = pattern.format(*name_parts)
        try:
            # Identify the first name/path in the registry that matches
            name = next(k for k in POOCH.registry.keys() if k.endswith(name))
        except StopIteration:
            # Invalid name, not in the registry
            continue

        # Try to fetch if `fetch` is True; else return every name
        yield Path(POOCH.fetch(name) if fetch else name)


def _make_geo_codes(
    cl: "sdmx.model.common.Codelist", values: pd.Series, **kwargs
) -> dict:
    """Create a codelist for the ``GEO`` concept, given certain `values`.

    For each unique value in `values`:

    1. Strip leading and trailing whitespace.
    2. Pass the data through :data`.transport_data.util.pycountry.NAME_MAP` for commonly
       used non-ISO values.
    3. Look up the value in ISO 3166-1 via :any:`pycountry.countries`.
    4. Add a code to the returned code list. If the lookup in (3) succeeds, the code ID
       is the alpha-2 code, and the name and description are populated from the
       database. If the lookup in (3) fails, the code ID is a unique integer.

    Returns
    -------
    .Codelist
        One entry for each unique value in `values`.
    dict
        Mapping from `values` to codes in the code list. Passing this to
        :meth:`pandas.Series.replace` should convert the `values` to the corresponding
        codes.
    """
    from pycountry import countries
    from sdmx.model import v21

    from transport_data.util.pycountry import NAME_MAP

    counter = count()
    id_for_name: dict[str, str] = {}

    @lru_cache
    def _make_code(value: str):
        name = value.strip()
        try:
            # - Apply replacements from NAME_MAP.
            # - Lookup in ISO 3166-1.
            match = countries.lookup(NAME_MAP.get(name.lower(), name))
        except LookupError:
            try:
                # Look up an already-generated code that matches this `value`
                code = cl[id_for_name[value]]
            except KeyError:
                # Generate a new code with a serial ID
                code = v21.Code(id=str(next(counter)), name=name)
        else:
            # Found an entry in the ISO 3166-1 database; duplicate it into the codelist
            code = v21.Code(
                id=match.alpha_2,
                name=match.name,
                description="Identical to the ISO 3166-1 entry",
            )

        try:
            cl.append(code)  # Add to the code list
        except ValueError:
            pass  # Already exists

        # Update the map
        id_for_name.setdefault(value, code.id)

    values.sort_values().drop_duplicates().apply(_make_code)

    return id_for_name


@hookimpl
def get_agencies():
    """Return the OICA Agency."""
    from sdmx.model import v21

    a = v21.Agency(
        id="OICA",
        name="International Organization of Motor Vehicle Manufacturers",
        description="https://www.oica.net",
    )
    return (a,)


@hookimpl
def provides():
    return (
        "Codelist=TDCI:CL_OICA_GEO",
        "ConceptScheme=TDCI:CS_OICA_CONCEPTS",
    )


def get_cl_geo() -> "sdmx.model.common.Codelist":
    """Retrieve or create the ``Codelist=TDCI:OICA_GEO``."""
    import sdmx.urn
    from sdmx.model import common

    from transport_data import STORE, org

    candidate: common.Codelist = common.Codelist(
        id=f"{get_agencies()[0].id}_GEO",
        maintainer=org.get_agencies()[0],
    )
    assert candidate.maintainer is not None

    # Check for an existing version of this object
    if versions := STORE.list_versions(
        common.Codelist, candidate.maintainer.id, candidate.id
    ):
        # ≥1 version exists; get and return it
        candidate.version = versions[-1]
        return STORE.get(sdmx.urn.make(candidate))
    else:
        # No existing artefact; store the candidate with a default version and return it
        candidate.version = "1.0.0"
        STORE.set(candidate)
        return candidate


@lru_cache
def get_conceptscheme() -> "sdmx.model.common.ConceptScheme":
    """Return a concept scheme with OICA concepts."""
    from sdmx.model import common

    from transport_data import STORE, org

    cs = common.ConceptScheme(
        id=f"{get_agencies()[0].id}_CONCEPTS",
        maintainer=org.get_agencies()[0],
        version="0.1",
    )

    # Measures
    for id_, name in (
        ("PROD", "Production"),
        ("SALES", "Sales"),
        ("SALES_GR", "Sales growth over a given period"),
        ("STOCK", "Vehicles in use"),
        ("STOCK_AAGR", "Vehicles in use, average annual growth rate"),
        ("STOCK_CAP", "Vehicles in use per capita"),
    ):
        cs.append(common.Concept(id=id_, name=name))

    # TODO replace the following with references to a common TDCI concept scheme
    # Used as dimensions
    for id_, core_rep in (
        ("GEO", get_cl_geo()),
        ("VEHICLE_TYPE", None),
        ("TIME_PERIOD", None),
    ):
        c = common.Concept(id=id_)
        c.core_representation = (
            common.Representation(enumerated=core_rep) if core_rep else None
        )
        cs.append(c)
    # Used as attributes
    for id_ in ("UNIT_MEASURE",):
        cs.append(common.Concept(id=id_))
    # Used as primary measure
    for id_ in ("OBS_VALUE",):
        cs.append(common.Concept(id=id_))

    STORE.set(cs)

    return cs


def get_structures(
    measure: str,
) -> tuple[
    "sdmx.model.v21.DataflowDefinition", "sdmx.model.v21.DataStructureDefinition"
]:
    """Create a data flow and data structure definitions for a given OICA `measure`.

    The DFD and DSD have URNs like ``TDCI:DF_OICA_{MEASURE}(0.1)``. The DSD has:

    - dimensions ``GEO``, ``VEHICLE_TYPE``, ``TIME_PERIOD``,
    - attributes ``UNIT_MEASURE``, and
    - primary measure with ID ``OBS_VALUE``.
    """
    from sdmx.model import v21

    from transport_data import STORE, org

    base = f"{get_agencies()[0].id}_{measure}"
    ma_args = dict(
        maintainer=org.get_agencies()[0],
        version="0.1",
        is_final=False,
        is_external_reference=False,
    )

    dfd = v21.DataflowDefinition(id=f"DF_{base}", **ma_args)
    STORE.set(dfd)
    dsd = v21.DataStructureDefinition(id=f"DS_{base}", **ma_args)
    STORE.set(dsd)

    if not dfd.is_final:
        # Associate
        dfd.structure = dsd

        # Mark as complete
        dfd.is_final = True
        STORE.set(dfd)

    if not dsd.is_final:
        cs = get_conceptscheme()

        for d, local_rep in (
            ("GEO", get_cl_geo()),
            ("VEHICLE_TYPE", None),
            ("TIME_PERIOD", None),
        ):
            dsd.dimensions.getdefault(
                id=d,
                concept_identity=cs[d],
                local_representation=v21.Representation(enumerated=local_rep),
            )

        for a in ("UNIT_MEASURE",):
            dsd.attributes.getdefault(id=a, concept_identity=cs[d])

        for m in ("OBS_VALUE",):
            dsd.measures.getdefault(id=m, concept_identity=cs[d])

        # Mark as complete
        dsd.is_final = True
        STORE.set(dsd)

    return dfd, dsd


def update_registry() -> None:
    """Update the registry.

    This function crawls :data:`.BASE_URL` for file names matching certain patterns.
    Files that exist are downloaded, hashed, and added to the :data:`.REGISTRY_FILE`.

    As of 2024-02-05, about 40 distinct files can be retrieved with :data:`.POOCH`. To
    view the list of files, use:

    .. code-block:: python

       from transport_data import oica
       print(oica.POOCH.registry)
    """
    from pooch import file_hash

    for dfd, _ in map(get_structures, ["PROD", "SALES", "STOCK"]):
        for file in filenames_for_dfd(dfd, fetch=False):
            filename = file.name
            existing_hash = POOCH.registry.setdefault(filename, None)

            if not POOCH.is_available(filename):
                # File doesn't exist on the remote
                POOCH.registry.pop(filename)
                continue

            if existing_hash is None:  # pragma: no cover
                # Download the file to add its hash
                path = POOCH.fetch(filename)
                POOCH.registry[filename] = file_hash(path)

    with open(REGISTRY_FILE, "w") as f:
        for k, v in sorted(POOCH.registry.items()):
            f.write(f"{k}  {v}\n")



"""Utilities for :mod:`pooch`."""

import logging
import time
from collections.abc import Callable
from functools import cache, partial
from http import HTTPStatus
from pathlib import Path

import pooch
import pooch.downloaders
from pooch.downloaders import (
    DEFAULT_TIMEOUT,
    DataRepository,
    DataverseRepository,
    FigshareRepository,
    HTTPDownloader,
    choose_downloader,
)
from pooch.utils import parse_url

log = logging.getLogger(__name__)


class DOIDownloader(pooch.downloaders.DOIDownloader):
    def __call__(self, url: str, output_file, pooch: "Pooch") -> None:
        from requests.exceptions import HTTPError

        # NB Identical to the upstream version, except use the Pooch._repository
        #    attribute. This avoids repeatedly querying the API.
        data_repository = pooch._repository
        assert data_repository is not None

        # Resolve the URL; remove a leading slash in the path
        file_name = parse_url(url)["path"].lstrip("/")
        download_url = data_repository.download_url(file_name)

        # Possibly append request headers, e.g. with a Zenodo API token
        kwargs = self.kwargs.copy()
        if pooch._repository and hasattr(pooch._repository, "get_headers"):
            kwargs.setdefault("headers", {})
            kwargs["headers"] |= pooch._repository.get_headers()

        # Instantiate the downloader object
        downloader = HTTPDownloader(
            progressbar=self.progressbar, chunk_size=self.chunk_size, **kwargs
        )
        retry(
            partial(downloader, download_url, output_file, pooch),
            lambda r: not isinstance(r, HTTPError),
        )


class Pooch(pooch.Pooch):
    """:class:`pooch.Pooch` with special powers.

    Parameters
    ----------
    module : str
        The `path` argument to pooch.Pooch is derived automatically from this, within
        the transport-data directory.
    expand : callable, optional
        If given, a function that receives any argument and returns a file name found
        within the Pooch registry.
    processor :
        If the exact value "unzip", fetch() operations are always processed to unzip
        archive contents into the same directory as the downloaded file.
    """

    _downloader_cls: type
    _processor = None
    _repository: "DataRepository | None" = None

    def __init__(
        self,
        # Args of pooch.Pooch.__init__()
        base_url,
        registry: dict | Path | None = None,
        urls: dict | None = None,
        retry_if_failed: int = 0,
        allow_updates: bool = True,
        # Additional args
        module: str = "",
        expand: Callable | None = None,
        processor: str | None = None,
    ) -> None:
        kwargs = dict(
            path=pooch.os_cache("transport-data").joinpath(module.split(".")[1]),
            base_url=base_url,
            registry=None if isinstance(registry, Path) else registry,
            urls=urls,
            retry_if_failed=retry_if_failed,
            allow_updates=allow_updates,
        )

        # Identify the Downloader subclass that will be used
        self._downloader_cls = type(choose_downloader(base_url))
        # Use the modified DOIDownloader from the current module
        if issubclass(DOIDownloader, self._downloader_cls):
            self._downloader_cls = DOIDownloader

        # Store the callback
        self._expand = expand or str

        if processor == "unzip":
            self._processor = pooch.Unzip(extract_dir=kwargs["path"])

        super().__init__(**kwargs)

        # Load registry from a file at the given path
        if isinstance(registry, Path):
            self.load_registry(registry)

    # Override pooch.Pooch methods
    def fetch(self, fname: str, *args, **kwargs) -> str:
        self._ensure_doi_registry()

        fname = self._expand(fname)
        kwargs.setdefault(
            "downloader",
            self._downloader_cls(progressbar=kwargs.get("progressbar", False)),
        )
        kwargs.setdefault("processor", self._processor)
        return super().fetch(fname, *args, **kwargs)

    def is_available(self, fname, *args, **kwargs) -> bool:
        self._ensure_doi_registry()
        _fname = self._expand(fname)

        try:
            return super().is_available(_fname, *args, **kwargs)
        except NotImplementedError:
            if self._downloader_cls is DOIDownloader:
                # Override the behaviour of pooch.Pooch. The registry has been
                # downloaded from the upstream source, so assume the advertised files
                # are indeed available.
                return True
            else:  # pragma: no cover
                raise

    # Additional methods
    def _ensure_doi_registry(self) -> None:
        """Ensure that :attr:`registry` is populated for a :class:`DOIDownloader`.

        This version differs from the parent class in calling
        :meth:`.Pooch.load_registry_from_doi` only if the registry has not already been
        loaded.
        """
        self.load_registry_from_doi()

    def load_registry_from_doi(self) -> None:
        """Populate the registry using the data repository API.

        This version differs from the parent class in using :func:`doi_to_repository`
        from the current module.
        """
        if self._downloader_cls is not DOIDownloader:
            return
        elif self.registry:
            return

        if self._repository is None:
            # Create a repository instance
            doi = self.base_url.replace("doi:", "")
            self._repository = doi_to_repository(doi)

        # Call registry population for this repository
        return self._repository.populate_registry(self)

    def path_for(self, *args, **kwargs):
        """Return a filename and local cache path for the data file."""
        return self.path.joinpath(self._expand(*args, **kwargs))


def retry(func: Callable, check: Callable):
    """Retry `func` until :py:`check(func())` returns :any:`True`."""
    msg = f"Got HTTP {HTTPStatus.TOO_MANY_REQUESTS} Too Many Requests"

    # Some Fibonacci numbers for delays
    for wait in (1, 2, 3, 5, 8, 13, 21, 34, 55, 89):
        try:
            result = func()
        except Exception as e:
            # func() raised an exception; pass this to check()
            result = e

        if check(result):
            return result
        else:
            log.warning(f"{msg}; retrying in {wait} s")

            time.sleep(wait)

    raise RuntimeError(msg)


class ZenodoRepository(pooch.downloaders.ZenodoRepository):
    """Custom version of :class:`pooch.downloaders.ZenodoRepository`."""

    @property
    def api_response(self):
        """Cached API response from Zenodo.

        This version differs from the upstream version:

        1. The record ID for the Zenodo API is constructed without the "zenodo." prefix.
        2. A keyring secret with ID ``api-token-zenodo``, if any, is used for the HTTP
           “Authorization” header.
        3. The request is retried with increasing delays until it succeeds.
        """
        from requests.exceptions import JSONDecodeError

        if self._api_response is None:
            import requests

            # Separate the Zenodo record ID from the full URL
            record_id = self.archive_url.rpartition("/zenodo.")[2]
            assert record_id.isdigit()  # The ID is always numeric
            url = f"{self.base_api_url}/{record_id}"

            # Maybe retrieve some headers including the Zenodo API token
            headers = self.get_headers()

            response = retry(
                partial(requests.get, url, timeout=DEFAULT_TIMEOUT, headers=headers),
                lambda r: r.status_code != HTTPStatus.TOO_MANY_REQUESTS,
            )

            try:
                self._api_response = response.json()
            except JSONDecodeError:  # pragma: no cover
                raise RuntimeError(
                    f"Received {response} from {url} with headers:\n"
                    f"{response.headers}\nand body:\n{response.content}"
                )

        return self._api_response

    @classmethod
    @cache
    def get_headers(cls) -> dict[str, str]:
        """Return a set of HTTP headers including the Zenodo API token, if any."""
        try:
            import keyring
        except ImportError:  # pragma: no cover
            pass
        else:
            if token := keyring.get_password("transport-data", "api-token-zenodo"):
                log.info("Using Zenodo API token from keyring")
                return {"Authorization": f"Bearer {token}"}
        return {}


def doi_to_repository(doi: str) -> "DataRepository":
    """Instantiate a data repository instance from a given DOI.

    This version differs from the upstream version in using :func:`doi_to_url` and
    :class:`ZenodoRepository` from the current module.
    """
    doi = doi.rstrip("/")

    # Construct a direct URL to the archive/record from the DOI
    archive_url = doi_to_url(doi)

    # Try the converters one by one until one of them returns a URL
    for cls in (FigshareRepository, ZenodoRepository, DataverseRepository):
        if data_repository := cls.initialize(archive_url=archive_url, doi=doi):
            return data_repository

    raise ValueError(f"Cannot construct a repository for DOI:{doi}")


def doi_to_url(doi: str) -> str:
    """Follow a DOI link to resolve the URL of the archive.

    This version differs from the upstream version in only making a single request, to
    doi.org. The HTTP 302 (redirect) response has a "location" header that gives the
    resolved URL for the DOI; this is returned directly and not queried.
    """
    import requests

    response = requests.get(
        f"https://doi.org/{doi}", timeout=DEFAULT_TIMEOUT, allow_redirects=False
    )
    assert response.status_code == HTTPStatus.FOUND
    return response.headers["location"]



import os

import pytest

from transport_data.estat import get, list_flows


def test_list_flows():
    list_flows()


@pytest.mark.flaky(condition="GITHUB_ACTIONS" in os.environ, reruns=5)
@pytest.mark.parametrize("df_id", list_flows())
def test_get(df_id):
    get(df_id)
